import numpy as np
from scipy.spatial.distance import cdist
import trimesh
import trimesh.intersections
import trimesh.triangles
import slam.geodesics
import slam.utils as utils


def cortical_surface_profiling(mesh, rot_angle, r_step, max_samples):
    """
    Surface profiling for a given cortical surface.
    NOTE:
    This function returns 2D profiling coordinates directly instead of 3D.
    These 2D points are used to generate the feature Maps.
    :param mesh: trimesh object
        The cortical surface mesh.
    :param rot_angle: float
        Degree of rotation angle.
    :param r_step: float
        Length of sampling steps.
    :param max_samples:
        Maximum of samples in one profiles.
    :return: (N_vertex, N_p, N_s) float
        Profiles points in their 2D coordinates.
        (x, y) respectively
    """

    vert = mesh.vertices
    poly = mesh.faces.astype(np.int32)
    norm = mesh.vertex_normals

    # compute the geodesic map of cortical surface within the specified radius
    # NOTE: This needs some time
    area_radius = r_step * max_samples * 2
    area_geodist = slam.geodesics.local_gdist_matrix(mesh, area_radius)

    #
    profile_samples_x = []
    profile_samples_y = []
    length = len(vert)

    for i in range(length):
        # for every vertex, do the surface profiling.
        vert_i = vert[i]
        vert_norm_i = norm[i]

        # limit the intersection area into the area_radius (on Distmap) from
        # center
        vert_distmap = area_geodist[i].toarray()[0]
        area_geodist_v = np.where(vert_distmap > 0)[0]
        area_geodist_faces = vert2poly_indices(area_geodist_v, poly)
        intersect_mesh = mesh.submesh(np.array([area_geodist_faces]))[0]

        # randomly select initial direction of rotation R0
        init_rot_dir = np.array([1, 1, 1]) - vert_i

        # get the profile samplings on surface
        sam_prof = surface_profiling_vert(
            vert_i,
            vert_norm_i,
            init_rot_dir,
            rot_angle,
            r_step,
            max_samples,
            intersect_mesh,
        )

        # compute the 2D coordinates (x, y) for all profile points
        sample_x, sample_y = compute_profile_coord_x_y(
            sam_prof, vert[i], norm[i])

        profile_samples_x.append(sample_x)
        profile_samples_y.append(sample_y)

    return np.array(profile_samples_x), np.array(profile_samples_y)


def surface_profiling_vert(
    vertex, vert_norm, init_rot_dir, rot_angle, r_step, max_samples, mesh
):
    """
    Implement the profile sampling process for a given vertex.
    NOTE:
    For a given vertex,
        Number of profiles N_p = 360/theta
    For each profiles,
        Number of Sampling points N_s = max_samples
    :param vertex: (3,) float
        Target vertex (center vertex)
    :param vert_norm: (3,) float
        Vertex normal
    :param init_rot_dir: (3, )
        Initial direction of rotation R0
    :param rot_angle: (3,) float
        Degree of rotation angle
    :param r_step: float
        Length of sampling steps
    :param max_samples: int
        Maximum of samples in one profiles
    :param mesh: trimesh object
        Intersecting mesh
    :return: (N_p, N_s, 3) float
        Profiles points in 3D coordinates.
    """

    profile_list = []  # record all the samples x and y
    vertex = np.array(vertex)

    # project the dir_R0 onto the tangent plane
    rot_vec0 = utils.project_vector2tangent_plane(vert_norm, init_rot_dir)[0]
    round_angle = 360

    for i in range(int(round_angle / rot_angle)):
        # set the rotation directions
        rot_angle_alpha = (i * rot_angle) * 1.0 / 360 * 2 * np.pi
        rot_mat_alpha = utils.get_rotate_matrix(vert_norm, rot_angle_alpha)
        rot_vec_alpha = np.dot(rot_vec0, rot_mat_alpha)
        p_norm = np.cross(vert_norm, rot_vec_alpha)

        # Get the intersection lines
        # the lines contains the rotation direction and the reverse one.
        intersect_lines = trimesh.intersections.mesh_plane(
            mesh, p_norm, vertex)

        # Select the points in the direction of rotation vector
        points_i, _, _ = select_points_orientation(
            intersect_lines, rot_vec_alpha, vertex, vert_norm
        )

        # Calculate the samples of profiles
        points_profile = compute_profiles_sampling_points(
            points_i, vertex, max_samples, r_step
        )

        profile_list.append(points_profile)

    return np.array(profile_list)


def second_round_profiling_vert(
    vertex,
    vert_norm,
    init_rot_dir,
    rot_angle,
    r_step,
    max_samples,
    mesh,
    mesh_face_index,
):
    """
    Implement the profile sampling process to get the feature values of each
    profiling points.
    The function name comes from the description of the method in the article。
    Different from the surface_profiling_vert, the mesh_face_index is
    obligatory.
    :param vertex: (3,) float
        Target vertex (center vertex)
    :param vert_norm: (3,) float
        Vertex normal
    :param init_rot_dir: (3, )
        Initial direction of rotation R0
    :param rot_angle: (3,) float
        Degree of rotation angle
    :param r_step: float
        Length of sampling steps
    :param max_samples: int
        Maximum of samples in one profiles
    :param mesh: trimesh object
        Intersecting mesh
    :param mesh_face_index:
        Indices of polygons of mesh.
        Use to record which polygon the sampling points belongs to.
    :return:
        profile_points: (N_p, N_s, 3, 3) float
            For each profile points contain [p1, p2, sample_points],
            where p1, p2 are the points used to calculate the sampling points.
        profile_intersect_faces: ((N_p, N_s,) int
    """

    profile_points = []  # record all the profiling points and interpolation
    # points
    profile_intersect_faces = []  # record all the faces id that contain the
    # sample points
    vertex = np.array(vertex)

    # project the dir_R0 onto the tangent plane
    rot_vec0 = utils.project_vector2tangent_plane(vert_norm, init_rot_dir)[0]
    round_angle = 360

    for i in range(int(round_angle / rot_angle)):
        # set the rotation directions
        rot_angle_alpha = (i * rot_angle) * 1.0 / 360 * 2 * np.pi
        rot_mat_alpha = slam.utils.get_rotate_matrix(
            vert_norm, rot_angle_alpha)
        rot_vec_alpha = np.dot(rot_vec0, rot_mat_alpha)
        p_norm = np.cross(vert_norm, rot_vec_alpha)

        # Get the intersection lines
        # the lines contains the rotation direction and the reverse one.
        intersect_lines, faces = trimesh.intersections.mesh_plane(
            mesh, p_norm, vertex, return_faces=True
        )

        # get the global index of faces
        intersect_fm_index = mesh_face_index[faces]

        # Select the points in the direction of rotation vector
        orient_points_i, orient_p_id, ori_lines_id = select_points_orientation(
            intersect_lines, rot_vec_alpha, vertex, vert_norm
        )

        orient_face_id = intersect_fm_index[ori_lines_id]

        # Calculate the samples of profiles
        points_interp_profile, cor_faces_index \
            = compute_profiles_sampling_points(
                orient_points_i,
                vertex,
                max_samples,
                r_step,
                orient_face_id
                )

        profile_points.append(points_interp_profile)
        profile_intersect_faces.append(cor_faces_index)

    return np.array(profile_points), np.array(profile_intersect_faces)


def compute_profiles_sampling_points(
    points_intersect, origin, max_samples, r_step, face_id=None
):
    """
    Calculate the sampling points on each profiles.
    :param points_intersect: (n, 3) float
    :param origin: (3,) float
        origin vertex
    :param max_samples: int
        Maximum of samples in one profiles
    :param r_step: float
        Length of sampling steps
    :param face_id: (n,) int
        Indices of polygons which intersecting points belong to.
        Default is None, it is only used in the second round profiling.
    :return:
        When face_id is None, return
            sampling points on profiles: (n, 3) float
        Otherwise,
            points_interpolate_profile: (n, 3, 3) float
                contains [p1, p2, sample_points]
                where p1, p2 are the points used to calculate the sampling
                points.
            cor_faces_index: (n,)
                the corresponding faces of profile points
    """

    if len(points_intersect) == 0:
        # the origin point is out of intersection points
        profile_points = list(np.zeros([max_samples, 3]))
        return profile_points

    # record the length of segment
    length_sum = np.linalg.norm(points_intersect[0] - origin)
    minued_lenth = 0
    count_i = 0
    # record the i when the sample distance firstly exceed the maximum
    # of intersection
    exceed_index = 0
    exceed_bool = True
    count_max = len(points_intersect)

    profile_points = []
    # Record the two end-points and sampling points
    points_interpolate_profile = []
    # the corresponding faces
    cor_faces_index = []

    for j in range(max_samples):
        sample_dist = (j + 1) * r_step
        if sample_dist <= length_sum and count_i == 0:
            # the point is between center point and the closest one
            point0 = origin
            point1 = points_intersect[0]

        elif sample_dist <= length_sum and count_i != 0:
            point0 = points_intersect[count_i - 1]
            point1 = points_intersect[count_i]

        else:
            minued_lenth = length_sum
            count_i += 1

            # the distance of sample exceed the local maximum
            if count_i == count_max:
                # first time arrive at boundary
                if exceed_bool:
                    exceed_index = j
                    exceed_bool = False
                count_i -= 1
                sample_dist = (exceed_index + 1) * r_step
                point0 = points_intersect[count_i - 1]
                point1 = points_intersect[count_i]
            else:
                point0 = points_intersect[count_i - 1]
                point1 = points_intersect[count_i]
                length_sum += np.linalg.norm(point1 - point0)

        if np.linalg.norm(point1 - point0) == 0:
            alpha = 0
        else:
            alpha = (sample_dist - minued_lenth) / \
                np.linalg.norm(point1 - point0)

        sample_point = (1 - alpha) * point0 + alpha * point1

        profile_points.append(sample_point)
        points_interpolate_profile.append([point0, point1, sample_point])
        # save the related intersect mesh faces
        if face_id is not None:
            cor_faces_index.append(face_id[count_i])

    if face_id is None:
        return profile_points

    return points_interpolate_profile, cor_faces_index


def vert2poly_indices(vertex_array, poly_array):
    """
    Find vertex-polygon indices from the polygons array of vertices
    TODO There is a func in the lastest trimesh:
        trimesh.geometry.vertex_face_indices()
    :param vertex_array:
    :param poly_array:
    :return:
    """

    vert_poly_arr = np.array([], dtype=int)
    for i in range(len(vertex_array)):
        poly_i = np.where(poly_array == vertex_array[i])[0]
        vert_poly_arr = np.hstack((vert_poly_arr, poly_i))

    return np.unique(vert_poly_arr)


def select_points_orientation(intersect_points, r_alpha, origin, norm):
    """
    Select points in a specified orientation,
    and ordered them by distance from the center.
    :param intersect_points: (n, 2, 3) float
        Points of intersecting lines.
    :param r_alpha: (3,) float
        Orientation vector
    :param origin: (3,) float
        Origin point
    :param norm: (3,) float
        Normal of origin point
    :return: orient_points,    (n, 3) float
             orient_p_indices, (n,) int
             lines_indices,    (n,) int
        Ordered points in the orientation.
    """

    points_i = intersect_points.reshape(intersect_points.size // 3, 3)

    # find the center points
    p_idx, count_coord = np.unique(
        np.where(points_i == origin)[0], return_counts=True)
    origin_index = p_idx[np.where(count_coord == 3)[0]]

    if len(origin_index) == 0:
        # the intersection result exclude origin point
        orient_points = []
        orient_p_indices = []
        lines_indices = []
        return orient_points, orient_p_indices, lines_indices

    ordered_points, ordered_p_indices = radial_sort(points_i, origin, norm)

    orientation_vec = np.dot(ordered_points - origin, r_alpha)
    orient_point_idx = np.where(orientation_vec > 0)[0][::2]
    orient_points = ordered_points[orient_point_idx]
    orient_p_indices = ordered_p_indices[orient_point_idx]

    # find the closest point
    p2o_len = cdist(orient_points, np.array([origin]), metric="euclidean")
    p2o_len = p2o_len.reshape(p2o_len.shape[0])
    ordered_p2o = np.argsort(p2o_len)
    orient_points = orient_points[ordered_p2o]
    orient_p_indices = orient_p_indices[ordered_p2o]

    # get the ordered intersection lines
    ori_lines_indices = orient_p_indices // 2

    return orient_points, orient_p_indices, ori_lines_indices


def radial_sort(points, origin, normal):
    """
    NOTE:
    This function is derived from the
        trimesh.points.radial_sort(points_i, origin, norm)
    I overwrite this function to return both the coordinates and indices of
    the points.
    Sorts a set of points radially (by angle) around an
    an axis specified by origin and normal vector.
    Parameters
    --------------
    points : (n, 3) float
      Points in space
    origin : (3,)  float
      Origin to sort around
    normal : (3,)  float
      Vector to sort around
    Returns
    --------------
    ordered : (n, 3) float
      Same as input points but reordered
    """

    # create two axis perpendicular to each other and the normal,
    # and project the points onto them
    axis0 = [normal[0], normal[2], -normal[1]]
    axis1 = np.cross(normal, axis0)
    pt_vec = points - origin
    pr0 = np.dot(pt_vec, axis0)
    pr1 = np.dot(pt_vec, axis1)

    # calculate the angles of the points on the axis
    angles = np.arctan2(pr0, pr1)

    # return the points and their indices sorted by angle
    return points[(np.argsort(angles))], np.argsort(angles)


def compute_profile_coord_x_y(profile, origin, normal):
    """
    Calculate the 2D coordinates of the profiling points in their rotation
    planes
    These points are used to generate the feature maps mentioned in the
    articles.
    :param profile: (N_p, N_s, 3) float
        Sampling points of profiles in 3D
    :param origin: (3,) float
        Center vertex of profiles
    :param normal: (3,) float
        Normal of origin
    :return: (N_p, N_s) float
        The coordinate x, y
    """
    num_prof = len(profile)
    num_sample = len(profile[0])

    # flat the samples
    profile_samples = profile.reshape([num_prof * num_sample, 3])

    # get the vector
    pro_sam_vec = profile_samples - origin
    vec_x = utils.project_vector2tangent_plane(normal, pro_sam_vec)
    vec_y = utils.project_vector2vector(normal, pro_sam_vec)

    # the length of the vector that projected onto the normal vector
    # x
    length_x = np.linalg.norm(vec_x, axis=1)
    x = length_x.reshape([num_prof, num_sample])

    # y
    sign_y = np.sign((profile_samples - origin).dot(normal))
    length_y = np.linalg.norm(vec_y, axis=1) * sign_y
    y = length_y.reshape([num_prof, num_sample])

    return x, y


def get_texture_value_on_profile(
    texture, mesh, profiling_samples, profiling_samples_fid
):
    """
    Calculate the texture values of each points on profiles by barycentric
    interpolation
    :param texture: slam texture
    :param mesh: trimesh object
    :param profiling_samples: (N, N_p, N_s, 3, 3) float
        N = Number of center vertices for surface profiling.
        N_p = Number of profiles for each center.
        N_s = Number of sampling points on each profiles.
        3 = [p1, p2, sampling points]
        3 = (3,) float
    :param profiling_samples_fid: (N, N_p, N_s) int
        Faces id corresponding to the profile sampling points.
    :return:
        texture_profile: (N, N_p, N_s) float
    """

    # compute the barycentric parameters of each profile point to its
    # co-faces of mesh
    barycentric_coord = compute_profile_barycentric_para(
        profiling_samples, mesh, profiling_samples_fid
    )

    # compute the features of each profile
    tex_arr = texture.darray[0]
    texture_profile = compute_profile_texture_barycentric(
        tex_arr, mesh, profiling_samples_fid, barycentric_coord
    )

    return texture_profile


def compute_profile_barycentric_para(profile_sample_points, mesh, triangle_id):
    """
    Compute the barycentric parameters of each points on profiles
    :param profile_sample_points: (N, N_p, N_s, 3, 3) float
        N = Number of center vertices for surface profiling.
        N_p = Number of profiles for each center.
        N_s = Number of sampling points on each profiles.
        3 = [p1, p2, sampling points]
        3 = (3,) float
    :param mesh: trimesh object
    :param triangle_id: (N, N_p, N_s) int
        Faces id corresponding to the profile sampling points.
    :return:
        barycentric: (N, N_p, N_s, 3) float
        Barycentric coordinates for all profiles points
    """

    if len(profile_sample_points.shape) != 5:
        raise Exception(
            "Wrong type of profile_sample_points, "
            "it must be (N, N_p, N_s, 3, 3)."
        )

    vert = mesh.vertices
    poly = mesh.faces

    # get the sample points on profile
    sample_points_profile = profile_sample_points[:, :, :, 2]
    sample_points = sample_points_profile.reshape(
        sample_points_profile.size // 3, 3)

    # get the faces
    triangle_id = triangle_id.reshape(triangle_id.size)
    triangles_v = vert[poly[triangle_id]]

    barycentric = trimesh.triangles.points_to_barycentric(
        triangles_v, sample_points)
    barycentric = barycentric.reshape(
        len(sample_points_profile),
        len(sample_points_profile[0]),
        len(sample_points_profile[0][0]),
        3,
    )

    return barycentric


def compute_profile_texture_barycentric(
        texture, mesh, triangle_id, barycentric_coord):
    """
    Compute the texture values of each points on profiles
    :param texture: darray of slam texture
    :param mesh: trimesh object
    :param triangle_id: (N, N_p, N_s) int
        Faces id corresponding to the profile sampling points.
    :param barycentric_coord: (N, N_p, N_s, 3) float
         Barycentric coordinates for all profiles points
    :return:
    """

    num_profiles = len(barycentric_coord)
    num_areas = len(barycentric_coord[0])
    num_sides = len(barycentric_coord[0][0])
    poly = mesh.faces

    triangle_id = triangle_id.reshape(triangle_id.size)
    barycentric = barycentric_coord.reshape(barycentric_coord.size // 3, 3)

    feature_tri_points = texture[poly[triangle_id]]
    texture_profile = np.dot(feature_tri_points * barycentric, [1, 1, 1])

    return texture_profile.reshape(num_profiles, num_areas, num_sides)



import numpy as np
from trimesh import util as tut

# from trimesh.geometry import mean_vertex_normals
import slam.topology as stop


def norm(vector):
    return np.sqrt(np.sum(vector**2))


def curvature_fit(mesh, tol=1e-12, neighbour_size=2):
    """
    Computation of the two principal curvatures based on:
    Petitjean, A survey of methods for recovering quadrics
    in triangle meshes, ACM Computing Surveys, 2002
    :param mesh:
    :param tol:
    :param neighbour_size:
    :return:
    """
    N = mesh.vertices.shape[0]
    # vertex_normals = mean_vertex_normals(N,
    # faces=mesh.faces, face_normals = mesh.face_normals)
    # vertex_normals, Avertex, Acorner, up, vp =
    # calcvertex_normals(mesh, mesh.face_normals)
    vertex_normals = mesh.vertex_normals

    curvature = np.zeros((N, 2))
    directions = np.zeros((N, 3, 2))

    adjacency_matrix = stop.adjacency_matrix(mesh)

    for i in range(N):
        if i == 175:
            print("Debug")
        # Definition of local basis
        point = np.reshape(mesh.vertices[i, :], (3, 1))
        normal = vertex_normals[i, :].transpose()
        normal = normal / norm(normal)
        normal = np.reshape(normal, (3, 1))
        vec1, vec2 = determine_local_basis(normal, tol)
        # rotation_matrix = np.concatenate((vec1.transpose(),
        #                                  np.reshape(vec2, (1, 3)),
        #                                  normal.transpose()))
        rotation_matrix = np.concatenate(
            (vec1, vec2, normal), axis=1).transpose()

        # neighbours
        neigh = stop.k_ring_neighborhood(
            mesh, index=i, k=neighbour_size, adja=adjacency_matrix
        )
        # neigh =
        # np.logical_and(neigh <= neighbour_size, neigh > 0).nonzero()[0]
        neigh = (neigh <= neighbour_size).nonzero()[0]
        neigh_len = len(neigh)
        vertices_neigh = mesh.vertices[neigh, :].transpose()
        # translation, origin at point
        vertices_neigh = vertices_neigh - np.repeat(point, neigh_len, axis=1)
        rotated_vertices_neigh = np.matmul(rotation_matrix, vertices_neigh)

        # Parametric model
        X = np.reshape(rotated_vertices_neigh[0, :], (neigh_len, 1))
        Y = np.reshape(rotated_vertices_neigh[1, :], (neigh_len, 1))
        Z = np.reshape(rotated_vertices_neigh[2, :], (neigh_len, 1))
        XY = np.concatenate((X**2, X * Y, Y**2), axis=1)
        parameters = np.matmul(np.linalg.pinv(XY), Z)
        # parameters = np.linalg.lstsq(XY,Z)

        # Curvature tensor
        tensor = np.array(
            [
                [parameters[0, 0], parameters[1, 0] / 2],
                [parameters[1, 0] / 2, parameters[2, 0]],
            ]
        )
        eigval, eigvec = np.linalg.eig(tensor)
        curvature[i, 0] = 2 * eigval[0]
        curvature[i, 1] = 2 * eigval[1]
        directions[i, :, 0] = np.matmul(
            rotation_matrix[0:2, :].transpose(), eigvec[:, 0]
        ).transpose()
        directions[i, :, 1] = np.matmul(
            rotation_matrix[0:2, :].transpose(), eigvec[:, 1]
        ).transpose()

    # Sort
    indices = np.argsort(curvature, axis=1)
    curvature = np.sort(curvature, axis=1)
    indices = np.expand_dims(indices, axis=1)
    directions = np.take_along_axis(directions, indices, axis=2)

    return curvature, directions


def determine_local_basis(normal, tol):
    """Compute an orthonormal basis of a 2d plane defined by its normal.

    A test vector (1,0,0) is first projected on the 2d plane.
    The normalized projection vec1 constituting the first vector of the basis.
    The second vector vec2 is obtained using cross product.

    :param normal: unitary normal vector to the plan
    :type normal: (3,1) ndarray
    :param tol: minimal norm value
    :type: np.float64
    :return:  vec1, vec2 two orthonormal vectors generating the plan
    :rtype: (2,) tuple of ndarray each of size (3,1)
    """
    # (3,3) projection matrix on the tangent plane
    proj_matrix = np.identity(3) - np.matmul(normal, normal.transpose())
    # project (1,0,0) orthogonaly on the plane
    vec1 = np.matmul(proj_matrix, np.array([[1], [0], [0]]))
    # vec1 vanishes if the normal is close to (1,0,0)
    if np.abs(norm(vec1)) < tol:
        vec1 = np.matmul(proj_matrix, np.array([[0], [1], [0]]))
    vec1 = vec1 / norm(vec1)
    vec2 = np.cross(normal[:, 0], vec1[:, 0])
    vec2 = vec2 / norm(vec2)
    vec2 = np.reshape(vec2, (3, 1))
    return vec1, vec2


def project_curvature_tensor(uf, vf, nf, old_ku, old_kuv, old_kv, up, vp):
    """
    performs a projection of the tensor variables to the vertex
    coordinate system
    :param uf:face coordinates system
    :param vf:face coordinates system
    :param nf:face normal
    :param old_ku:face curvature tensor variables
    :param old_kuv:face curvature tensor variables
    :param old_kv:face curvature tensor variables
    :param up:vertex cordinate system
    :param vp:vertex cordinate system
    :return:
    new_ku,new_kuv,new_kv : vertex curvature tensor coordinates
    The tensor : [[new_ku, new_kuv], [new_kuv, new_kv]]
    """

    r_new_u, r_new_v = rotate_coordinate_system(up, vp, nf)
    OldTensor = np.array([[old_ku, old_kuv], [old_kuv, old_kv]])
    u1, v1 = np.dot(r_new_u, uf), np.dot(r_new_u, vf)
    u2, v2 = np.dot(r_new_v, uf), np.dot(r_new_v, vf)

    new_ku = (u1 * (OldTensor[0, 0] * u1 + OldTensor[0, 1] * v1)
              + v1 * (OldTensor[1, 0] * u1 + OldTensor[1, 1] * v1))
    new_kuv = (u1 * (OldTensor[0, 0] * u2 + OldTensor[0, 1] * v2)
               + v1 * (OldTensor[1, 0] * u2 + OldTensor[1, 1] * v2))
    new_kv = (u2 * (OldTensor[0, 0] * u2 + OldTensor[0, 1] * v2)
              + v2 * (OldTensor[1, 0] * u2 + OldTensor[1, 1] * v2))

    return new_ku, new_kuv, new_kv


def compute_curvature(FV, VertexNormals, FaceNormals,
                      Avertex, Acorner, up, vp):
    """
    CalcFaceCurvature recives a list of vertices and faces in FV structure
    and the normal at each vertex and calculates the second fundemental
    matrix and the curvature using least squares
    :param FV: face-vertex data structure containing a list of vertices and a
    list of faces
    :param VertexNormals: n*3 matrix ( n = number of vertices ) containing the
    normal at each vertex
    :param FaceNormals: m*3 matrix ( m = number of faces ) containing the
    normal of each face
    :param Avertex:
    :param Acorner:
    :param up:
    :param vp:
    :return:
    FaceSFM - an m*1 cell matrix second fundemental
    VertexSFM - an n*w cell matrix second fundementel
    wfp - corner voronoi weights
    """
    print("Calculating curvature tensors ... Please wait")
    "Matrix of each face at each cell"
    FaceSFM = np.zeros((FV.faces.shape[0], 2, 2))
    VertexSFM = np.zeros((FV.vertices.shape[0], 2, 2))
    Kn = np.zeros((1, FV.faces.shape[0]))

    " Get all the edge vectors "
    e0 = FV.vertices[FV.faces[:, 2], :] - FV.vertices[FV.faces[:, 1], :]
    e1 = FV.vertices[FV.faces[:, 0], :] - FV.vertices[FV.faces[:, 2], :]
    e2 = FV.vertices[FV.faces[:, 1], :] - FV.vertices[FV.faces[:, 0], :]

    " Normalize edge vectors "
    e0_norm = tut.unitize(e0)
    # e1_norm = normr(e1)
    # e2_norm = normr(e2)

    wfp = np.array(np.zeros((FV.faces.shape[0], 3)))
    for i in range(FV.faces.shape[0]):
        "Calculate Curvature Per Face"
        "set face coordinate frame"
        nf = FaceNormals[i, :]
        t = e0_norm[i, :]
        B = np.cross(nf, t)
        B = B / (np.linalg.norm(B))

        "extract relevant normals in face vertices"
        n0 = VertexNormals[FV.faces[i][0], :]
        n1 = VertexNormals[FV.faces[i][1], :]
        n2 = VertexNormals[FV.faces[i][2], :]

        " solve least squares problem of th form Ax=b "
        A = np.array(
            [
                [np.dot(e0[i, :], t), np.dot(e0[i, :], B), 0],
                [0, np.dot(e0[i, :], t), np.dot(e0[i, :], B)],
                [np.dot(e1[i, :], t), np.dot(e1[i, :], B), 0],
                [0, np.dot(e1[i, :], t), np.dot(e1[i, :], B)],
                [np.dot(e2[i, :], t), np.dot(e2[i, :], B), 0],
                [0, np.dot(e2[i, :], t), np.dot(e2[i, :], B)],
            ]
        )

        b = np.array(
            [
                np.dot(n2 - n1, t),
                np.dot(n2 - n1, B),
                np.dot(n0 - n2, t),
                np.dot(n0 - n2, B),
                np.dot(n1 - n0, t),
                np.dot(n1 - n0, B),
            ]
        )

        "Resolving by least mean square method because "
        "A is not a square matrix"

        x = np.linalg.lstsq(A, b, None)

        FaceSFM[i] = np.array([[x[0][0], x[0][1]], [x[0][1], x[0][2]]])
        Kn[0][i] = np.dot(
            np.array([1, 0]), np.dot(FaceSFM[i], np.array([[1.0], [0.0]]))
        )
        """
        Calculate curvature per vertex
        Calculate voronoi weights
        """
        wfp[i][0] = Acorner[i][0] / Avertex[FV.faces[i][0]]
        wfp[i][1] = Acorner[i][1] / Avertex[FV.faces[i][1]]
        wfp[i][2] = Acorner[i][2] / Avertex[FV.faces[i][2]]

        "Calculate new coordinate system and project the tensor"

        for j in range(3):
            new_ku, new_kuv, new_kv = project_curvature_tensor(
                t,
                B,
                nf,
                x[0][0],
                x[0][1],
                x[0][2],
                up[FV.faces[i][j], :],
                vp[FV.faces[i][j], :],
            )
            VertexSFM[FV.faces[i][j]] += np.dot(
                wfp[i][j], np.array([[new_ku, new_kuv], [new_kuv, new_kv]])
            )

    print("Finished Calculating curvature tensors")

    return FaceSFM, VertexSFM, wfp


def curvatures_and_derivatives(mesh):
    (VertexNormals, Avertex, Acorner, up, vp) = calcvertex_normals(
        mesh, mesh.face_normals
    )
    (FaceSFM, VertexSFM, wfp) = compute_curvature(
        mesh, VertexNormals, mesh.face_normals, Avertex, Acorner, up, vp
    )
    [PrincipalCurvature, PrincipalDi1, PrincipalDi2] = principal_curvatures(
        mesh, VertexSFM, up, vp
    )
    return PrincipalCurvature, PrincipalDi1, PrincipalDi2


# def calcfacenormals(FV):
#     """
#     Calculates face normals
#     :param FV: face vertex data structure containing a list of vertices and a
#     list of faces
#     :return:
#     Face normals : matrix that contains each faces' normal
#     (dim = number of faces * 3 )
#     """
#
#     "Get all edge vectors"
#     e0 = FV.vertices[FV.faces[:, 2], :] - FV.vertices[FV.faces[:, 1], :]
#     e1 = FV.vertices[FV.faces[:, 0], :] - FV.vertices[FV.faces[:, 2], :]
#
#     "Calculate and return normal of face"
#     return normr(np.cross(e0, e1))


# def normr(X):
#     """
#     Returns a matrix with the same size where each row normalized to a
#     vector length of 1
#     :param X:
#     :return:
#     """
#
#     if len(np.shape(X)) == 1:
#         return X / np.abs(X)
#     else:
#         a = np.shape(X)[1]
#         b = np.shape(X)[0]
#         return np.dot(np.reshape(np.transpose(
#             np.sqrt(1 / somme_colonnes(np.transpose(X ** 2)))), (b, 1)),
#             np.ones((1, a))) * X


def calcvertex_normals(FV, N):
    """
    CalcVertexNormals calculates the normals and voronoi areas at each vertex
    INPUT:
    FV - triangle mesh in face vertex structure
    N - face normals
    OUTPUT -
    VertexNormals - [Nv X 3] matrix of normals at each vertex
    Avertex - [NvX1] voronoi area at each vertex
    Acorner - [NfX3] slice of the voronoi area at each face corner
    """

    print("Calculating vertex normals .... Please wait")

    "Get all the edge vectors"
    e0 = np.array(FV.vertices[FV.faces[:, 2], :] -
                  FV.vertices[FV.faces[:, 1], :])
    e1 = np.array(FV.vertices[FV.faces[:, 0], :] -
                  FV.vertices[FV.faces[:, 2], :])
    e2 = np.array(FV.vertices[FV.faces[:, 1], :] -
                  FV.vertices[FV.faces[:, 0], :])

    "Normalize edge vectors "
    e0_norm = tut.unitize(e0)
    e1_norm = tut.unitize(e1)
    e2_norm = tut.unitize(e2)

    de0 = np.sqrt((e0[:, 0]) ** 2 + (e0[:, 1]) ** 2 + (e0[:, 2]) ** 2)
    de1 = np.sqrt((e1[:, 0]) ** 2 + (e1[:, 1]) ** 2 + (e1[:, 2]) ** 2)
    de2 = np.sqrt((e2[:, 0]) ** 2 + (e2[:, 1]) ** 2 + (e2[:, 2]) ** 2)
    l2 = np.array([de0**2, de1**2, de2**2])
    l2 = np.transpose(l2)
    """
    using ew to compute the cot of the angles for the voronoi area calculation
    ew is the triangle barycenter.
    We check later if it's inside or outside the triangle
    """
    ew = np.array(
        [
            l2[:, 0] * (l2[:, 1] + l2[:, 2] - l2[:, 0]),
            l2[:, 1] * (l2[:, 2] + l2[:, 0] - l2[:, 1]),
            l2[:, 2] * (l2[:, 0] + l2[:, 1] - l2[:, 2]),
        ]
    )
    s = (de0 + de1 + de2) / 2

    "Af - face area vector"
    Af = np.sqrt(s * (s - de0) * (s - de1) * (s - de2))

    "herons formula for triangle area, "
    "could have also used  0.5 * norm(cross(e0,e1)) "
    "Calc weights"
    Acorner = np.zeros((np.shape(FV.faces)[0], 3))
    Avertex = np.zeros((np.shape(FV.vertices)[0], 1))

    "Compute vertex normals"
    VertexNormals, up, vp = (
        np.zeros((np.shape(FV.vertices)[0], 3)),
        np.zeros((np.shape(FV.vertices)[0], 3)),
        np.zeros((np.shape(FV.vertices)[0], 3)),
    )

    for i in range(np.shape(FV.faces)[0]):
        wfv1 = Af[i] / ((de1[i] ** 2) * (de2[i] ** 2))
        wfv2 = Af[i] / ((de0[i] ** 2) * (de2[i] ** 2))
        wfv3 = Af[i] / ((de1[i] ** 2) * (de0[i] ** 2))

        VertexNormals[FV.faces[i][0], :] += wfv1 * N[i, :]
        VertexNormals[FV.faces[i][1], :] += wfv2 * N[i, :]
        VertexNormals[FV.faces[i][2], :] += wfv3 * N[i, :]

        """
        Calculate areas for weights according to Mayar et al. [2002]
        Check if the triangle is obtuse, right or acute
        """
        "Changed shape for ew"

        if ew[0][i] <= 0:
            Acorner[i][1] = (
                -0.25 * l2[i][2] * Af[i] /
                (np.dot(e0[i, :], np.transpose(e2[i, :])))
            )
            Acorner[i][2] = (
                -0.25 * l2[i][1] * Af[i] /
                (np.dot(e0[i, :], np.transpose(e1[i, :])))
            )
            Acorner[i][0] = Af[i] - Acorner[i][2] - Acorner[i][1]
        elif ew[1][i] <= 0:
            Acorner[i][2] = (
                -0.25 * l2[i][0] * Af[i] /
                (np.dot(e1[i, :], np.transpose(e0[i, :])))
            )
            Acorner[i][0] = (
                -0.25 * l2[i][2] * Af[i] /
                (np.dot(e1[i, :], np.transpose(e2[i, :])))
            )
            Acorner[i][1] = Af[i] - Acorner[i][2] - Acorner[i][0]
        elif ew[2][i] <= 0:
            Acorner[i][0] = (
                -0.25 * l2[i][1] * Af[i] /
                (np.dot(e2[i, :], np.transpose(e1[i, :])))
            )
            Acorner[i][1] = (
                -0.25 * l2[i][0] * Af[i] /
                (np.dot(e2[i, :], np.transpose(e0[i, :])))
            )
            Acorner[i][2] = Af[i] - Acorner[i][1] - Acorner[i][0]
        else:
            ewscale = 0.5 * Af[i] / (ew[0][i] + ew[1][i] + ew[2][i])
            Acorner[i][0] = ewscale * (ew[1][i] + ew[2][i])
            Acorner[i][1] = ewscale * (ew[0][i] + ew[2][i])
            Acorner[i][2] = ewscale * (ew[1][i] + ew[0][i])

        Avertex[FV.faces[i][0]] += Acorner[i][0]
        Avertex[FV.faces[i][1]] += Acorner[i][1]
        Avertex[FV.faces[i][2]] += Acorner[i][2]

        " Calcul initial coordinate system "
        up[FV.faces[i][0], :] = e2_norm[i, :]
        up[FV.faces[i][1], :] = e0_norm[i, :]
        up[FV.faces[i][2], :] = e1_norm[i, :]

    VertexNormals = tut.unitize(VertexNormals)

    " Calcul initial vertex coordinate system"

    for i in range(np.shape(FV.vertices)[0]):
        up[i, :] = np.cross(up[i, :], VertexNormals[i, :])
        up[i, :] = up[i, :] / np.linalg.norm(up[i, :])
        vp[i, :] = np.cross(VertexNormals[i, :], up[i, :])

    print("Finished calculating vertex normals")
    return VertexNormals, Avertex, Acorner, up, vp


def principal_curvatures(FV, VertexSFM, up, vp):
    """
    Calculates the principal curvatures and prncipal directions
    :param FV: triangular mesh
    :param VertexSFM: second fundemental matrix for each vertex
    :param up: vertex local coordinate frame
    :param vp: vertex local coordinate frame
    :return:
    PrincipalCurvature : Matrix containing pricipale curvatures
    ( dim = 2 * Number of vertices )
    PrincipalDi1 , PrincipalDi2 : First and second principal directions
    """

    print("Calculating Principal Components ... Please wait")

    "Calculate principal curvatures"
    PrincipalCurvature = np.zeros((2, np.shape(FV.vertices)[0]))
    PrincipalDi1, PrincipalDi2 = [
        np.zeros((np.shape(FV.vertices)[0], 3)),
        np.zeros((np.shape(FV.vertices)[0], 3)),
    ]
    for i in range(np.shape(FV.vertices)[0]):
        npp = np.cross(up[i, :], vp[i, :])
        r_old_u, r_old_v = rotate_coordinate_system(up[i, :], vp[i, :], npp)
        ku = VertexSFM[i][0][0]
        kuv = VertexSFM[i][0][1]
        kv = VertexSFM[i][1][1]
        c, s, tt = 1, 0, 0
        if kuv != 0:
            "Jacobi rotation to diagonalize"
            h = 0.5 * (kv - ku) / kuv
            if h < 0:
                tt = 1 / (h - np.sqrt(1 + h**2))
            else:
                tt = 1 / (h + np.sqrt(1 + h**2))
            c = 1 / np.sqrt(1 + tt**2)
            s = tt * c
        k1 = ku - tt * kuv
        k2 = kv + tt * kuv
        if abs(k1) >= abs(k2):
            PrincipalDi1[i, :] = c * r_old_u - s * r_old_v
        else:
            [k1, k2] = [k2, k1]
            PrincipalDi1[i, :] = s * r_old_u + c * r_old_v
        PrincipalDi2[i, :] = np.cross(npp, PrincipalDi1[i, :])
        PrincipalCurvature[0][i] = k1
        PrincipalCurvature[1][i] = k2

        if np.isnan(k1) or np.isnan(k2):
            print("Nan")

    print("Finished Calculating principal components")
    return PrincipalCurvature, PrincipalDi1, PrincipalDi2


def rotate_coordinate_system(up, vp, nf):
    """
    RotateCoordinateSystem performs the rotation of the vectors up and vp
    to the plane defined by nf as its normal vector
    :param up: vectors to be rotated (vertex coordinate system)
    :param vp: vectors to be rotated (vertex coordinate system)
    :param nf: face normal
    :return:
    r_new_u,r_new_v : new rotated vectors
    """

    r_new_u = up
    r_new_v = vp
    npp = np.cross(up, vp) / np.linalg.norm(np.cross(up, vp))
    ndot = np.dot(nf, np.transpose(npp))
    if ndot <= -1:
        r_new_u = -r_new_u
        r_new_v = -r_new_v
    perp = nf - ndot * npp
    dperp = (npp + nf) / (1 + ndot)
    r_new_u = r_new_u - dperp * np.dot(perp, np.transpose(r_new_u))
    r_new_v = r_new_v - dperp * np.dot(perp, np.transpose(r_new_v))

    return r_new_u, r_new_v


def decompose_curvature(in_curv):
    """
    Decompose the principal curvatures of a mesh in ShapeIndex and curvedness
    Implementation based on
    ' Surface shape and curvature scales
      Jan JKoenderink & Andrea Jvan Doorn
      Image and Vision Computing
      Volume 10, Issue 8, October 1992, Pages 557-564 '
    """
    curvatures = np.array(
        (np.maximum(in_curv[0], in_curv[1]),
         np.minimum(in_curv[0], in_curv[1]))
    )
    shapeIndex = (2 / np.pi) * np.arctan(
        (curvatures[0, :] + curvatures[1, :]) /
        (curvatures[1, :] - curvatures[0, :])
    )
    curvedness = np.sqrt((curvatures[0, :] ** 2 + curvatures[1, :] ** 2) / 2)
    return shapeIndex, curvedness


def curvedness_shapeIndex(mesh):
    curv = curvatures_and_derivatives(mesh)[0]
    return decompose_curvature(curv)


# def somme_colonnes(X):
#     """
#
#     INPUT :
#     X : A matrix with any dimension
#     OUTPUT:
#     A row that contains the sum of each column
#     """
#     xx = list()
#     for i in range(np.shape(X)[1]):
#         xx.append(sum(X[:, i]))
#     return np.array(xx)



import unittest

import numpy as np

import slam.differential_geometry as sdg
import slam.generate_parametric_surfaces as sgpm

TOL = 1e-15
TOL2 = 1e-1


def cartesian_to_spherical(coords):
    n = len(coords)
    spherical_coordinates = np.zeros((n, 3))
    for i in range(n):
        R = np.sqrt(np.sum(coords[i, :] ** 2))
        spherical_coordinates[i, 0] = R
        coord_i = coords[i, :] / R
        spherical_coordinates[i, 2] = np.arctan2(
            coord_i[1], coord_i[0]
        )  # Phi, in [0,2pi]
        spherical_coordinates[i, 1] = np.arccos(coord_i[2])  # Theta, in [0,pi]
    return spherical_coordinates


class TestDifferentialGeometry(unittest.TestCase):
    # def test_gradient(self):
    #     # Trivial example at the moment:
    #     # Uniform texture: gradient vanishes
    #     mesh = trimesh.creation.icosphere(subdivisions=4, radius=1.0)
    #     n_vert = mesh.vertices.shape[0]
    #     uniform_texture = np.ones((n_vert,))
    #     gradient_uniform = sdg.gradient(mesh, uniform_texture)
    #     self.assertTrue(
    #         (np.abs(gradient_uniform - np.zeros((n_vert, 3))) < TOL).all())

    def test_gradient(self):
        sphere_mesh = sgpm.generate_sphere_icosahedron(subdivisions=4)
        spherical_coordinates = cartesian_to_spherical(sphere_mesh.vertices)
        phi = spherical_coordinates[:, 2]
        theta = spherical_coordinates[:, 1]
        gradient_theta = sdg.gradient_fast(sphere_mesh, theta)

        analytic_gradient_theta = np.zeros((sphere_mesh.vertices.shape[0], 3))
        analytic_gradient_theta[:, 0] = np.cos(theta) * np.cos(phi)
        analytic_gradient_theta[:, 1] = np.cos(theta) * np.sin(phi)
        analytic_gradient_theta[:, 2] = -np.sin(theta)
        error_norms = np.sqrt(
            np.sum((gradient_theta - analytic_gradient_theta) ** 2, axis=1)
        )
        # error_norms decrease linearly with mesh spacing
        self.assertTrue(np.mean(error_norms) < TOL2)


if __name__ == "__main__":
    unittest.main()



"""Spectral Analysis of Gyrification (SPANGY)


This module implements the Spectral Analysis framework described in [1]_ and
applied in [2]_. Hence this module:

1. Perform spectral decomposition of a mesh (eigenvalues and eigenfunctions
   of the laplace-beltrami operator defined on the mesh), i.e. obtaining a
   functional basis
2. Decompose a scalar function living on the mesh (e.g. mean
   curvature) in this functional basis
3. Compute the power spectrum associated to the obtained spectral decomposition
4. Group power spectrum into frequency bands of interest


References
----------
.. [1] D. Germanaud*, J. Lefevre*,R. Toro, C. Fischer, J.Dubois, L.
 Hertz-Pannier, J.F. Mangin, Larger is twistier:Spectral Analysis of
 Gyrification (SPANGY) applied to adult brain size polymorphism,
 Neuroimage, 63 (3), 1257-1272, 2012.

.. [2] D. Germanaud, J. Lefèvre, C. Fischer, M. Bintner, A. Curie, V.
 des Portes, S. Eliez, M. Elmaleh-Bergès, D. Lamblin, S. Passemard,
 G. Operto, M. Schaer, A. Verloes, R. Toro, J.F. Mangin, L. Hertz-Pannier,
 Simplified gyral pattern in severe developmental microcephalies?
 New insights from allometric modeling for spatial and spectral analysis
 of gyrification,NeuroImage,Volume 102, Part 2, 2014,
 Pages 317-331, ISSN 1053-8119,
"""

import numpy as np
from scipy.sparse.linalg import eigsh

import slam.differential_geometry as sdg


def eigenpairs(mesh, nb_eig):
    """Compute nb_eig eigen pairs (eigen value and associated eigenvector) of
    the Laplace-Beltrami operator defined on the input mesh

    The nb_eig smallest eigen pairs of the spectrum of the discrete
    Laplace-Beltrami operator.

    Parameters
    ----------
    mesh : Base.Trimesh

    nb_eig : Int
        number of eigenpairs to compute.
    Returns
    -------
    eig_val : Array of float
        eigenvalues computed.
    eig_vec : Array of float
        eigenvectors computed.
    Notes
    -----
    The nb_eig smallest of the Laplacian spectrum are computed using the
    shift-invert method of eighsh as LAPACK library is far more efficient at
    finding the largest eigenvalues of a matrix
    (see https://docs.scipy.org/doc/scipy/tutorial/arpack.html) for complete
    explanations
    """
    lap, lap_b = sdg.compute_mesh_laplacian(mesh, lap_type='fem')
    eig_val, eig_vec = eigsh(lap.tocsr(), nb_eig, M=lap_b.tocsr(),
                             sigma=1e-6, which='LM')
    return eig_val, eig_vec, lap_b.tocsr()


def spectrum(f2analyse, mass_matrix, eig_vec, eig_val):
    """Compute the (grouped) spectrum of a scalar function  in a functional
    basis

    The functional basis is derived from the eigenvectors of a laplace
    beltrami operator defined on the mesh. The spectrum function:
        1. Compute the full discrete spectrum of the scalar function by
        computing a scalar product between this function and the eigenvectors
        corrected by the mass matrix associated with the mesh
        2. Compute the number of bands based on eigen values
        3. Defined the extents of the different bands
        4. Compute the power associated with each defined spectral band

    Parameters
    ----------
    f2analyse : Array of floats
        function to analyze (mean curvature).
    mass_matrix : Array of floats
        Used in the discretization of the eigenvalue problem.
    eig_vec : Array of floats
        Eigenvectors.
    eig_val : Array of floats
        Eigenvalues.
    Returns
    -------
    grouped_spectrum : Array of floats
        power in each spectral band.
    group_indices : Array of ints
        Indices of spectral bands.
    coefficients : Array of floats
        Fourier coefficients of the input function f2analyse.
    nlevels : int
        Number of spectral bands into which the raw power
        spectrum will be compacted.
    """

    coefficients = f2analyse.dot(mass_matrix.transpose().dot(eig_vec))
    nlevels = int(0.5 * np.log(eig_val[-1]
                               / eig_val[1]) / np.log(2)) + 2  # nlevels = 7
    grouped_spectrum = np.zeros((nlevels, 1))
    # projection coef of the neurofunctional signal on the principal
    # eigenpair (brain shape)
    grouped_spectrum[0] = coefficients[0]**2
    group_indices = np.zeros((nlevels, 2), dtype=int)
    group_indices[0, :] = [0, 0]

    for k in range(nlevels - 2):
        indice = np.where(eig_val >= eig_val[1] * 2 ** (2 * (k)))
        group_indices[k + 1, 0] = indice[0][0]
        indice = np.where(eig_val <= eig_val[1] * 2 ** (2 * (k + 1)))
        group_indices[k + 1, 1] = indice[0][-1]
        grouped_spectrum[k + 1] = \
            np.sum(coefficients[
                   group_indices[k + 1, 0]:group_indices[k + 1, 1] + 1]**2)

    group_indices[-1, 0] = group_indices[-2, 1] + 1
    group_indices[-1, 1] = eig_val.size - 1
    grouped_spectrum[-1] = \
        np.sum(coefficients[
               group_indices[-1, 0]:group_indices[-1, 1]]**2)

    return grouped_spectrum.squeeze(), group_indices, coefficients, nlevels


def local_dominance_map(
        coefficients, f2analyse, nlevels, group_indices, eig_vec):
    """
    Parameters
    ----------
    coefficients : Array of floats
        Fourier coefficients of the input function f2analyse
    f2analyse : Array of floats
        Scalar function to analyze (e.g. mean curvature)
    nlevels : Array of ints
        number of spectral bands
    group_indices : Array of ints
        indices of spectral bands
    eig_vec : Array of floats
        eigenvectors (reversed order for computation and memory reasons)

    Returns
    -------
    loc_dom_band : Array of floats
        texture with the band contributing the most to f2analyse
    frecomposed : Array of floats
        recomposition of f2analyse in each frequency band
    """
    N = np.size(coefficients)

    frecomposed = np.zeros((len(f2analyse), nlevels - 1), dtype='object')
    eig_vec = np.flip(eig_vec, 1)

    # band by band recomposition
    for i in range(nlevels - 1):
        # levels_ii: number of frequency band wihin the compact Band i
        levels_i = np.arange(
            group_indices[i + 1, 0], group_indices[i + 1, 1] + 1)
        # np.array((number of vertices, number of levels_ii))
        f_ii = np.dot(eig_vec[:, N - levels_i - 1], coefficients[levels_i].T)
        frecomposed[:, i] = f_ii

    # locally dominant band
    loc_dom_band = np.zeros((f2analyse.shape))

    diff_recomposed = frecomposed[:, 0]
    diff_recomposed = np.concatenate(
        (np.expand_dims(
            diff_recomposed, axis=1), np.diff(
            frecomposed, axis=1)), axis=1)

    # sulci
    idx = np.argmin(diff_recomposed, axis=1)
    idx = idx + 1
    loc_dom_band[f2analyse <= 0] = idx[f2analyse <= 0] * (-1)

    # gyri
    idx = np.argmax(diff_recomposed, axis=1)
    idx = idx + 1
    loc_dom_band[f2analyse > 0] = idx[f2analyse > 0]

    return loc_dom_band, frecomposed



"""
Unit test for the spangy module
TODO: everything
"""
import unittest
import trimesh
import numpy as np
import slam.spangy as spgy
import slam.differential_geometry as sdg
# import slam.curvature as scurv


def make_sphere():
    """ Create a sphere"""
    mesh_a = trimesh.creation.icosphere(subdivisions=1, radius=1.0)
    return mesh_a


def make_rectangle():
    """ Create a rectangle"""
    mesh_b = trimesh.creation.box(extents=[1.0, 2.0, 0.5])
    return mesh_b


class TestSpangy(unittest.TestCase):

    # Sphere
    sphere_test = make_sphere()
    rectangle_test = make_rectangle()
    precision = 1e-6

    def test_spangy_eigenpairs(self):
        """
        Test that the eigenvalues generation is correct, for a simple case.
        We compare the obtained results to the ones obtained with the matlab
        code. Which we assume is correct.
        """
        mesh = self.rectangle_test.copy()

        # Compute the eigenpairs, for N = 5
        N = 5
        eigVal, eigVects, _ = spgy.eigenpairs(mesh, N)
        # Compute the mesh laplacian outside
        lap, lap_b = sdg.compute_mesh_laplacian(mesh, lap_type='fem')
        assert np.allclose(lap @ eigVects - eigVal * (lap_b @ eigVects),
                           np.zeros((mesh.vertices.shape[0], N)
                                    ), atol=self.precision)
        # Compared to matlab results (hardcoded here for the same rectangle)
        eigVal_matlab = [1.31874474854710e-15,
                         1.99766647468559,
                         5.37024944254656,
                         7.38658922693569,
                         17.0853208854239]
        assert np.allclose(eigVal, eigVal_matlab, atol=self.precision)

    def test_spangy_spectrum(self):
        """
        Test that the spectrum generation is correct, for a simple case.

        We compare the obtained results to the ones obtained with the
        matlab code.
        which we assume is correct.
        """
        """
        mesh = self.rectangle_test.copy()
        # Compute the eigenpairs, for N = 5
        N = 5
        eigVal, eigVects, lap_b = spgy.eigenpairs(mesh, N)

        # Compute the curvature
        PrincipalCurvatures, _, _ = scurv.curvatures_and_derivatives(mesh)
        mean_curv = 0.5 * \
            (PrincipalCurvatures[0, :] + PrincipalCurvatures[1, :])

        grouped_spectrum, group_indices, _ = spgy.spectrum(mean_curv, lap_b,
                                                           eigVects, eigVal)
        levels = len(group_indices)
        assert levels == 3

        # Compared to matlab results (hardcoded here for the same rectangle)
        grouped_spectrum_matlab = [5.333333, 1.85197e-30, 4.930e-32]
        assert np.allclose(grouped_spectrum, grouped_spectrum_matlab,
                           atol=self.precision)
        """

    def test_spangy_local_dominance_map(self):
        """
        Test the output of the local dominance maps for a simple case.

        We compare the obtained results to the ones obtained with the
        matlab code.
        which we assume is correct.
        """

        """
        Test with a rectangle not the best idea for this test
        suggestions?

        mesh = self.rectangle_test.copy()
        # Compute the eigenpairs, for N = 5
        N = 5
        eigVal, eigVects, lap_b = spgy.eigenpairs(mesh, N)

        # Compute the curvature
        PrincipalCurvatures, _, _ = scurv.curvatures_and_derivatives(mesh)
        mean_curv = 0.5 * (PrincipalCurvatures[0, :] +
                           PrincipalCurvatures[1, :])

        grouped_spectrum, group_indices, coefficients = spgy.spectrum(
                                                             mean_curv, lap_b,
                                                             eigVects, eigVal)
        levels = len(group_indices)
        # eigVects = np.flip(eigVects, 1)
        # Compute the local dominance map
        loc_dom_band, frecomposed = spgy.local_dominance_map(coefficients,
                                                             mean_curv,
                                                             levels,
                                                             group_indices,
                                                             eigVects)

        # Compared to matlab results (hardcoded here for the same rectangle)
        # results differ! (sign issue)
        loc_dom_band_matlab = [1, 1, 2, 2, 1, 1, 2, 2]
        frecomposed_matlab = np.array([[4.92517e-16, -2.23134e-16],
                                       [2.34773e-16, -5.4280402e-17],
                                       [-5.4702e-16, 5.42804e-17],
                                       [-8.6377e-16, 2.23134e-16],
                                       [1.18850e-15, 5.428040e-17],
                                       [9.0114e-16, 2.2313e-16],
                                       [-5.29881e-16, -2.2313e-16],
                                       [-8.7624e-16, -5.428e-17]])
        import pdb; pdb.set_trace()
        assert np.allclose(frecomposed, frecomposed_matlab,
                           atol=self.precision)
        assert np.allclose(loc_dom_band, loc_dom_band_matlab,
                           atol=self.precision)
        """


if __name__ == "__main__":
    unittest.main()



def angle_difference(mesh1, mesh2):
    """
    compute the difference of angle between mesh1 and mesh2
    for each of the three angles of each face
    :param mesh1:
    :param mesh2:
    :return:
    """

    return mesh1.face_angles - mesh2.face_angles


def area_difference(mesh1, mesh2):
    """
    Compare_mesh_angle

    Difference between the faces.

    :param mesh1
    :type trimesh

    :param mesh2
    :type trimesh

    :return
    """

    return mesh1.area_faces - mesh2.area_faces


def edge_length_difference(mesh1, mesh2):
    """
    compute the difference of the length of edges between mesh1 and mesh2
    :param mesh1:
    :param mesh2:
    :return:
    """

    return mesh1.edges_unique_length - mesh2.edges_unique_length



import numpy as np
import itertools


def vertex_voronoi(mesh):
    """
    Implemented by Guillaume Auzias, Maxime Dieudonné and Lucile Hashimoto
    Compute vertex voronoi of a mesh as described in
    Meyer, M., Desbrun, M., Schroder, P., Barr, A. (2002).
    Discrete differential geometry operators for triangulated 2manifolds.
    Visualization and Mathematics, 1..26.
    :param mesh: trimesh object
    :return: numpy array of shape (mesh.vertices.shape[0],)
    """
    Nbp = mesh.faces.shape[0]  # Number of polygon faces
    obt_angs = mesh.face_angles > np.pi / 2
    obt_poly = obt_angs[:, 0] | obt_angs[:, 1] | obt_angs[:, 2]
    print('    -percent polygon with obtuse angle ',
          100.0 * len(np.where(obt_poly)[0]) / Nbp)
    cot = 1 / np.tan(mesh.face_angles)
    #
    # Extract the segments of the faces (pairs of vertices)
    # Comb contains the vertex indices of the three segments
    # of each face (Nbp*3,2)
    comb = np.array(
        [list(itertools.combinations(mesh.faces[i], 2)) for i in range(Nbp)])
    comb = comb.reshape([comb.size // 2, 2])
    #
    # Extract the coordinates of the vertices of the segments:
    # V1 for first column, V2 for the second
    V1 = np.array([mesh.vertices[i] for i in comb[:, 0]])
    V2 = np.array([mesh.vertices[i] for i in comb[:, 1]])
    #
    # Dist array : compute the square of the norm of each segments.
    # We need to switch the first and the last colums to gets
    # the right order : D0, D1, D2
    # faces | segments of the faces
    # F0    | D0 D1 D2
    # F1    | D0 D1 D2
    # FN    | D0 D1 D2
    Dist = np.sum(np.power(V1 - V2, 2), axis=1)
    Dist = Dist.reshape([Dist.size // 3, 3])
    Dist[:, [2, 0]] = Dist[:, [0, 2]]
    #
    # VorA0 : compute the voronoi area (when the face triangle is Aigu)
    # for all the V0 of the faces
    # VorA1 : compute the voronoi area (when the face triangle is Aigu)
    # for all the V1 of the faces
    # VorA2 : compute the voronoi area (when: the face triangle is Aigu)
    # for all the V2 of the faces
    # VorA : concatenation of VorA0, VorA1, VorA2.
    # faces | segments of the faces
    # F0    | VorA0 VorA1 VorA2
    # F1    | VorA0 VorA1 VorA2
    # FN    | VorA0 VorA1 VorA2
    VorA0 = Dist[:, 1] * cot[:, 1] + Dist[:, 2] * cot[:, 2]
    VorA1 = Dist[:, 0] * cot[:, 0] + Dist[:, 2] * cot[:, 2]
    VorA2 = Dist[:, 0] * cot[:, 0] + Dist[:, 1] * cot[:, 1]
    VorA = np.array([VorA0, VorA1, VorA2]).transpose().flatten()

    face_flat = mesh.faces.flatten()
    area_face = mesh.area_faces
    area_face_df = np.repeat(area_face, 3)
    obt_poly_df = np.repeat(obt_poly, 3)
    obt_angs_flat = obt_angs.flatten()
    #
    # we create 3 arrays according the condition on Fobt and Aobt.
    # The 3 possibilities are :
    #  (Fobt, Aobt) = (False, False) -> the voronoi area of a vertex
    #  in 1 triangle face is given in VorA
    #  (Fobt, Aobt) = (True, False)  -> the voronoi area of a vertex
    #  in 1 triangle face is given by area_face/4
    #  (Fobt, Aobt) = (True, True)  -> the voronoi area of a vertex
    #  in 1 triangle face is given by area_face/2
    # area_VorA :
    # sum of the vornoi area only for vertex with angle in aigue faces
    # area_VorOA :
    # sum of the vornoi area only for vertex with Aigue angle in  Obtue faces
    # area_VorOO :
    # sum of the vornoi area only for vertex with Obtue angle in  Obtue faces
    # area_VorA
    mask_AAAF = ~obt_poly_df
    vertices = face_flat[mask_AAAF]
    VorA_values = VorA[mask_AAAF]
    # Replace "groupby" vertex and "sum" from pandas
    unique_vertices, inverse_indices = np.unique(vertices, return_inverse=True)
    area_VorA = np.zeros_like(unique_vertices, dtype=np.float64)
    np.add.at(area_VorA, inverse_indices, VorA_values)
    area_VorA /= 8
    area_VorA = np.column_stack((unique_vertices, area_VorA))
    # area_VorOA
    mask_AAOF = obt_poly_df & ~obt_angs_flat
    vertices = face_flat[mask_AAOF]
    area_values = area_face_df[mask_AAOF]
    unique_vertices, inverse_indices = np.unique(vertices, return_inverse=True)
    area_VorOA = np.zeros_like(unique_vertices, dtype=np.float64)
    np.add.at(area_VorOA, inverse_indices, area_values)
    area_VorOA /= 4
    area_VorOA = np.column_stack((unique_vertices, area_VorOA))
    # area_VorOO
    mask_OAOF = obt_poly_df & obt_angs_flat
    vertices = face_flat[mask_OAOF]
    area_values = area_face_df[mask_OAOF]
    unique_vertices, inverse_indices = np.unique(vertices, return_inverse=True)
    area_VorOO = np.zeros_like(unique_vertices, dtype=np.float64)
    np.add.at(area_VorOO, inverse_indices, area_values)
    area_VorOO /= 2
    area_VorOO = np.column_stack((unique_vertices, area_VorOO))
    #
    # Then the voronoi area of one vertex in the mesh is given by the sum
    # of all the vornoi area in its neighboored triangle faces
    # so we sum for each vertex the area given in area_VorA,area_VorOA and
    # area_VorOO if the value exist
    # area :
    # | idx Vertex      | VorA  | VorOA | VorOO |
    # | 1               |   0.2 |  1.2  | Nan |  -> sum = area_vor
    # | 2               |   0.7 |  Nan  | Nan |  -> sum
    # | ...             |
    # | 480 562         |                        -> sum
    areas = np.vstack((area_VorA, area_VorOA, area_VorOO))
    unique_vertices, inverse_indices = np.unique(areas[:, 0],
                                                 return_inverse=True)
    area_Vor = np.zeros_like(unique_vertices, dtype=np.float64)
    np.add.at(area_Vor, inverse_indices, areas[:, 1])

    return area_Vor



"""
github.com/gauzias/slam
----------------------------

definition of the Texture class
"""

import numpy as np
from scipy import stats as sps


class TextureND:
    def __init__(self, darray=None, process=True, metadata=None, **kwargs):
        """
        TODO: better description
        A TextureND object contains a darray of shape nb texture times the
         number of vertices of the mesh the texture corresponds to.

        Parameters
        ----------
        metadata : dict
          Any metadata about the mesh
        process : bool
          if True, Nan and Inf values will be removed

        """
        if darray is None:
            self.darray = None
            self.dtype = None
            self.shape = None
            self.is_empty = True
        else:
            # (n, 3) float, set of vertices
            self.darray = np.atleast_2d(np.asanyarray(darray))
            self.dtype = self.darray.dtype
            self.shape = self.darray.shape

        # store metadata about the texture in a dictionary
        self.metadata = dict()
        # update the mesh metadata with passed metadata

        if isinstance(metadata, dict):
            self.metadata.update(metadata)

        if self.shape is not None:
            self.is_empty = False

        # process will remove NaN and Inf values and merge vertices
        # if validate, will remove degenerate and duplicate faces
        if process:
            self.process()

        # save reference to kwargs
        self._kwargs = kwargs

    def process(self):
        """
        Do the bare minimum processing to make a mesh useful.

        Does this by:
            1) removing NaN and Inf values

            2) merging duplicate vertices

        If self._validate:
            3) Remove triangles which have one edge of their rectangular 2D
               oriented bounding box shorter than tol.merge

            4) remove duplicated triangles

        Returns
        ------------
        self: trimesh.Trimesh
          Current mesh
        """
        # if there are no vertices or faces exit early
        if self.is_empty:
            return self

        self.remove_infinite_values()
        self.metadata["processed"] = True

    def remove_infinite_values(self):
        """
        Ensure that every data_array consists of finite numbers.

        This will remove np.nan and np.inf
        """

        # if the texture is already empty we can't remove anything
        if self.is_empty:
            return

        ind_infinite = ~np.isfinite(self.darray)
        darray = self.darray.copy()
        darray[ind_infinite] = 0
        self.update_darray(darray)

    def update_darray(self, darray):
        """
        Update darray and dtype, shape accordingly.
        :param darray:
        :return: Current TextureND
        """

        self.darray = np.asanyarray(darray)
        self.dtype = self.darray.dtype
        self.shape = self.darray.shape
        if self.shape is not None:
            self.is_empty = False

    def copy(self):
        """
        Safely get a copy of the current texture.

        Returns
        ---------
        copied : TextureND
          Copy of current texture
        """
        copied = TextureND()
        copied.update_darray(self.darray)
        copied.metadata = self.metadata.copy()

        return copied

    def min(self):
        return self.darray.min()

    def max(self):
        return self.darray.max()

    def mean(self):
        return self.darray.mean()

    def z_score_filtering(self, z_thresh=3):
        """
        Filter out values in darray where z_score > z_thresh
        The original outlier value is replaced by max(darray[not outlier])
        :param z_thresh: z_score threshold
        :return:
        """
        print(self.darray.shape)
        filtered_darray = self.darray.copy()
        for ind, d in enumerate(self.darray):
            z = sps.zscore(d)
            outliers_pos = z > z_thresh
            outliers_neg = z < -z_thresh
            outliers = outliers_pos | outliers_neg
            replace_value_pos = np.max(d[~outliers])
            replace_value_neg = np.min(d[~outliers])
            filtered_darray[ind, outliers_pos] = replace_value_pos
            filtered_darray[ind, outliers_neg] = replace_value_neg
        self.darray = filtered_darray
        print(self.darray.shape)
        self.metadata["z_score_filtered"] = True
        self.metadata["z_score_threshold"] = z_thresh



import numpy as np
import nibabel as nb
import trimesh

from slam import texture


def load_mesh(gifti_file):
    """
    load gifti_file and create a trimesh object
    :param gifti_file: str, path to the gifti file on the disk
    :return: the corresponding trimesh object
    """
    g = nb.load(gifti_file)
    coords, faces = (
        g.get_arrays_from_intent(
            nb.nifti1.intent_codes["NIFTI_INTENT_POINTSET"])[0].data,
        g.get_arrays_from_intent(
            nb.nifti1.intent_codes["NIFTI_INTENT_TRIANGLE"])[0].data,
    )
    metadata = g.meta.metadata
    metadata["filename"] = gifti_file
    return trimesh.Trimesh(
        faces=faces, vertices=coords, metadata=metadata, process=False
    )


def write_mesh(mesh, gifti_file):
    """Create a mesh object from two arrays

    fixme:  intent should be set !
    """
    coord = mesh.vertices
    triangles = mesh.faces
    carray = nb.gifti.GiftiDataArray(
        coord.astype(
            np.float32),
        "NIFTI_INTENT_POINTSET")
    tarray = nb.gifti.GiftiDataArray(
        triangles.astype(np.float32), "NIFTI_INTENT_TRIANGLE"
    )
    img = nb.gifti.GiftiImage(darrays=[carray, tarray])
    # , meta=mesh.metadata)

    nb.save(img, gifti_file)


def load_texture(gifti_file):
    """
    load gifti_file and create a TextureND object (multidimensional)
    :param gifti_file: str, path to the gifti file on the disk
    :return: the corresponding TextureND object
    """
    # read the gifti usinng nibabel
    # nb_texture = nb.gifti.read(gifti_file)
    nb_texture = nb.load(gifti_file)
    # concatenate all the data arrays in a single numpy array
    cat_darrays = [nb_texture.darrays[i].data for i in range(
        len(nb_texture.darrays))]
    return texture.TextureND(
        darray=np.array(cat_darrays), metadata=nb_texture.meta.metadata
    )


def write_texture(tex, gifti_file):
    """
    TODO manage metadata
    write a TextureND object to disk as a gifti file
    :param gifti_file: str, path to the gifti file on the disk
    :return: the corresponding TextureND object
    """
    darrays_list = []
    for d in tex.darray:
        gdarray = nb.gifti.GiftiDataArray(d.astype(np.float32), 0)
        # gdarray.metadata = tex.metadata
        # print(gdarray.metadata)
        darrays_list.append(gdarray)
    out_texture_gii = nb.gifti.GiftiImage(darrays=darrays_list)
    # out_metadata = tex.metadata
    # print(out_metadata)
    # out_metadata['Name']=gifti_file
    # print(out_metadata)
    # out_texture_gii.set_metadata(nb.gifti.GiftiMetaData(out_metadata))
    # , meta=str(tex.metadata))

    nb.save(out_texture_gii, gifti_file)



import numpy as np
from scipy import sparse
import networkx as nx

import trimesh
from trimesh import graph
from trimesh import grouping


def compute_face_normal(face_vertices):
    if face_vertices.ndim < 3:
        face_vertices = np.array([face_vertices])
    vectors = face_vertices[:, 1:, :] - face_vertices[:, :2, :]

    cprod = np.cross(vectors[:, 0], vectors[:, 1])
    norm = np.sqrt(np.dot(cprod * cprod, [1.0] * cprod.shape[1]))
    # in-place reciprocal of nonzero norms
    norm **= -1
    # multiply by reciprocal of norm
    unit_cprod = cprod * norm.reshape((-1, 1))
    return unit_cprod


def boundary_angles(boundary, vertices_coord):
    """
    compute the angle between consecutive edges in a path
    :param boundary: list of vertex indices corresponding to the path
    :param vertices_coord: array (n, 3) of vertex coordinates
    :return:
    """
    # pp = coordinates of consecutive vertices, including last-to-first to
    # close the path
    pp = np.zeros((len(boundary), 3))
    pp[:-1, :] = vertices_coord[boundary[:-1], :] - \
        vertices_coord[boundary[1:], :]
    pp[-1, :] = vertices_coord[boundary[-1], :] - \
        vertices_coord[boundary[0], :]
    norm = np.sqrt(np.dot(pp**2, [1] * pp.shape[1]))
    # tile reciprocal of norm
    i_norm = norm**-1
    tiled = np.tile(i_norm, (pp.shape[1], 1)).T
    # multiply by reciprocal of norm
    u_pp = pp * tiled
    # u_pp is the unit vectors corresponding to the edges
    # u_qq is a copy of u_pp with a shift in indices of 1, so that u_pp and
    # u_qq are unit vectors of two consecutive edges
    u_qq = pp.copy()
    u_qq[1:, :] = u_pp[:-1, :]
    u_qq[0, :] = u_pp[-1, :]
    # dot product and angle computation
    dots = np.dot(u_pp * u_qq, [1.0] * u_pp.shape[1])
    ang = np.pi - np.arccos(np.clip(dots, -1, 1))
    ang = ang * 180 / np.pi
    return ang, norm


def boundaries_intersection(boundaries):
    """
    compute the intersections inside a boundary
    :param boundaries: list of list of vertex indices corresponding to the path
    :return: list of common vertices between each tuple
    """
    bound_conn = []
    for bound_ind1 in range(len(boundaries) - 1):
        for bound_ind2 in range(bound_ind1 + 1, len(boundaries)):
            common = set(boundaries[bound_ind1]).intersection(
                set(boundaries[bound_ind2])
            )
            if common:
                bound_conn.append([bound_ind1, bound_ind2, list(common)])
    return bound_conn


def close_mesh(mesh, boundary_in=None):
    """
    see also trimesh.repair.fill_holes(mesh):
    This function implements the following paper with slight modifications.
    Zhao, W., Gao, S., & Lin, H. (2007). A robust hole-filling algorithm for
     triangular mesh. The Visual Computer, 23(12), 987–997.
     https://doi.org/10.1007/s00371-007-0167-y
    :param mesh:
    :param boundary_in:
    :return:
    """
    if boundary_in is None:
        boundary_in = mesh_boundary(mesh)
    if len(boundary_in) == 0:  # no hole in the mesh, nothing to do!
        nb_verts_added = []
        return mesh, nb_verts_added

    """ copy mesh arrays to avoid caching from trimesh"""
    vertices = np.asanyarray(mesh.vertices)
    faces = np.asanyarray(mesh.faces)
    face_normals = np.asanyarray(mesh.face_normals)

    nb_verts_added = 0
    for bound_i in boundary_in:
        bound = bound_i.copy()
        bound.reverse()
        """ closing the hole with front propagation """
        f_b = ismember(faces, bound)
        boundIndsF = np.where(f_b[:, 0] | f_b[:, 1] | f_b[:, 1])[0]
        [bound_ang, edge_length] = boundary_angles(bound, vertices)
        r_min = min(edge_length)
        peaks = list()
        skip_angle_test = False
        while len(bound) > 2:
            nb_verts = 0
            admissible = True
            l_verts = vertices.shape[0]
            lb = len(bound) - 1
            # print(lb)
            # l_faces = faces.shape[0]
            # print(l_verts, l_faces)
            # print(bound_ang)
            bound_ang_m = min(bound_ang)
            ind_m = np.argmin(bound_ang)
            if bound_ang_m == np.inf:
                # only problematic angles around the boundary!
                # print('warning :: forced closing')
                [bound_ang, edge_length] = boundary_angles(bound, vertices)
                ind_m = np.argmin(bound_ang)
                bound_ang_m = 0
                skip_angle_test = True

            """ creating faces """
            if bound_ang_m < 75:
                if ind_m == 0:
                    add_faces = [bound[ind_m], bound[ind_m + 1], bound[lb]]
                elif ind_m == lb:
                    add_faces = [bound[ind_m], bound[0], bound[ind_m - 1]]
                else:
                    add_faces = [bound[ind_m],
                                 bound[ind_m + 1], bound[ind_m - 1]]
                nb_faces = 1

                K1 = ismember(faces[boundIndsF, :], add_faces[0])
                K2 = ismember(faces[boundIndsF, :], add_faces[1])
                K3 = ismember(faces[boundIndsF, :], add_faces[2])

                Ks = K1 | K2 | K3
                indsF = Ks[:, 0] & Ks[:, 1] & Ks[:, 2]
                if any(indsF):  # this face already exists!!!!
                    admissible = False
                    bound_ang[ind_m] = np.inf
                    peaks.append(bound[ind_m])
                else:
                    if len(bound) < 4:
                        add_faces = bound
                    else:
                        Ks = K2 | K3
                        indsF = np.sum(Ks, 1)
                        if max(indsF) > 1:  # the new edge already exists!
                            admissible = False
                            bound_ang[ind_m] = np.inf
            else:  # if bound_ang_m<135
                if ind_m == 0:
                    tri_verts = [bound[ind_m], bound[ind_m + 1], bound[lb]]
                    add_faces = [
                        [bound[ind_m], bound[ind_m + 1], l_verts],
                        [bound[ind_m], l_verts, bound[lb]],
                    ]
                elif ind_m == lb:
                    tri_verts = [bound[ind_m], bound[0], bound[ind_m - 1]]
                    add_faces = [
                        [bound[ind_m], bound[0], l_verts],
                        [bound[ind_m], l_verts, bound[ind_m - 1]],
                    ]
                else:
                    tri_verts = [bound[ind_m],
                                 bound[ind_m + 1], bound[ind_m - 1]]
                    add_faces = [
                        [bound[ind_m], bound[ind_m + 1], l_verts],
                        [bound[ind_m], l_verts, bound[ind_m - 1]],
                    ]
                nb_faces = 2
                nb_verts = 1
                if bound_ang_m < 135:
                    [add_verts, r] = create_vertex(vertices[tri_verts, :])
                else:  # bound_ang>135
                    # print('big angle!')
                    [add_verts, r] = create_vertex(
                        vertices[tri_verts, :], r=r_min)
                comp_faces = list(
                    set(np.unique(faces[boundIndsF, :])
                        ).difference(set(tri_verts))
                )
                comp_verts = vertices[comp_faces, :]
                dist_comp_verts = comp_verts - np.tile(
                    add_verts, (comp_verts.shape[0], 1)
                )
                """
                testing if the created vertex is not too close to another
                existing vertex
                """
                if min(np.sum(dist_comp_verts**2, 1) - r_min**2) < 0:
                    admissible = False
                    bound_ang[ind_m] = 0  # => close with one single face

            if admissible:
                if skip_angle_test:
                    teta = np.inf
                else:
                    """
                    test angle between normals,
                    if length(add_faces(:,1))>1 test only the first face as
                    they are in the same plane
                    """
                    norm_add_faces = np.zeros(3)
                    if nb_faces > 1:
                        pp = vertices[add_faces[0][1], :] - \
                            vertices[add_faces[0][0], :]
                        qq = add_verts - vertices[add_faces[0][0], :]
                    else:
                        pp = vertices[add_faces[1], :]
                        -vertices[add_faces[0], :]
                        qq = vertices[add_faces[2], :]
                        -vertices[add_faces[0], :]
                    norm_add_faces[0] = pp[1] * qq[2] - pp[2] * qq[1]
                    norm_add_faces[1] = pp[2] * qq[0] - pp[0] * qq[2]
                    norm_add_faces[2] = pp[0] * qq[1] - pp[1] * qq[0]
                    norm = np.sqrt(np.sum(norm_add_faces**2))
                    norm_add_faces = norm_add_faces / norm

                    f_b = ismember(faces[boundIndsF, :], bound[ind_m])
                    boundsVertIndsF = boundIndsF[f_b[:, 0]
                                                 | f_b[:, 1] | f_b[:, 2]]
                    boundary_normalf = face_normals[boundsVertIndsF, :]
                    teta = max(np.dot(boundary_normalf, norm_add_faces))
                """
                teta-np.cos(np.pi/2) <=> acos()>pi/2 and np.cos(np.pi/2)=0
                """
                if teta < 0:
                    peaks.append(bound[ind_m])
                    bound_ang[ind_m] = np.inf
                else:
                    skip_angle_test = False
                    if nb_verts > 0:
                        vertices = np.vstack((vertices, add_verts))
                        nb_verts_added = nb_verts_added + 1
                    faces = np.vstack((faces, add_faces))
                    # print(np.array(add_faces))
                    # print(vertices[np.array(add_faces)])
                    # add_normals, valid = trimesh.triangles.normals(
                    #     vertices[np.array(add_faces)]
                    # )
                    add_normals\
                        = compute_face_normal(vertices[np.array(add_faces)])

                    face_normals = np.vstack((face_normals, add_normals))
                    """ update of the boundary,  boundIndsF and bound_ang"""
                    if nb_faces < 2:
                        bound.pop(ind_m)
                    else:  # nb_faces==2
                        bound[ind_m] = l_verts
                    f_b = ismember(faces, bound)
                    boundIndsF = np.where(f_b[:, 0] | f_b[:, 1] | f_b[:, 2])[0]
                    [bound_ang, edge_length] = boundary_angles(bound, vertices)

    """ barycentric smoothing of the new vertices """
    if peaks is not None:
        smoothed_peak = local_barycentric_smoothing(faces, vertices, peaks)
        vertices[peaks, :] = smoothed_peak
    # creating the output mesh, note that face normals are also recomputed here
    closed_mesh = trimesh.Trimesh(
        faces=faces, vertices=vertices, metadata=mesh.metadata, process=False
    )

    return closed_mesh, nb_verts_added


def local_barycentric_smoothing(faces, vertices, vertices_to_smooth):
    """
    local smoothing of the coordinates of vertices_to_smooth
    the smoothed coordinates of each vertex in vertices_to_smooth in vertices
    corrrespond to the barycenter of its neigbhors
    :param faces:
    :param vertices:
    :param vertices_to_smooth:
    :return:
    """
    smoothed_vertices = np.zeros((len(vertices_to_smooth), 3))
    for p, peak in enumerate(vertices_to_smooth):
        f_b = ismember(faces, peak)
        neigs = np.unique(faces[f_b[:, 0] | f_b[:, 1] | f_b[:, 2], :])
        smoothed_vertices[p, :] = np.mean(vertices[neigs, :], 0)
    return smoothed_vertices


def create_vertex(three_vertices, r=None):
    """
    create a vertex along the bisector of the angle formed by three vertices
    :param three_vertices: array(3,3) coordinates of the three vertices
    :param r: distance to the new vertex on the bisector
    :return:
    """
    # the two vectors formed by the three vertices
    pp = three_vertices[1, :] - three_vertices[0, :]
    qq = three_vertices[2, :] - three_vertices[0, :]
    # their norm
    n_pp = np.sqrt(np.sum(pp**2))
    n_qq = np.sqrt(np.sum(qq**2))
    # the two vectors normalized
    u_pp = pp / n_pp
    u_qq = qq / n_qq

    if r is None:
        r = min(n_pp, n_qq)
    # compute the bisector direction
    bisec = u_qq + u_pp
    """
    ensure the distance between the new vertex and three_vertices[0,:]
    is equal to r
    """
    k = np.sqrt(r**2 * np.sum(bisec**2)) / np.sum(bisec**2)
    new_vert = k * bisec + three_vertices[0, :]
    return new_vert, r


def cut_mesh(mesh, atex_in):
    """
    cut the mesh into submeshes following the values in the texture
    returns as many meshes as the number of different values in the texture
    the vertices on the border between two submeshes are duplicated into
    both submeshes
    :param mesh:
    :param atex_in:
    :return:
    """
    # see mesh.submesh(faces_sequence)
    atex = np.around(atex_in)
    atex2 = atex.copy()
    labels = np.unique(atex)
    labels = labels.tolist()
    labels.reverse()
    sub_meshes = list()
    sub_indexes = list()
    last_label = labels[-1]
    for label_ind in range(len(labels) - 1):
        (sub_mesh, sub_index) = sub_cut_mesh(mesh, atex, labels[label_ind])
        sub_meshes.append(sub_mesh)
        sub_indexes.append(sub_index.tolist())
        # boundary = texture_boundary(mesh, atex, labels[label_ind])
        boundary = texture_boundary_vertices(
            atex, labels[label_ind], mesh.vertex_neighbors
        )
        atex2[boundary] = last_label
    (sub_mesh, sub_index) = sub_cut_mesh(mesh, atex2, last_label)
    sub_meshes.append(sub_mesh)
    sub_indexes.append(sub_index.tolist())
    return sub_meshes, labels, sub_indexes


def edges_to_boundary(edges_bound, mesh_edges):
    """
    build the boundary by traversing edges return list of connected components
    ORDERED ACCORDING TO THEIR LENGTH, i.e. THE FIRST THE SHORTEST
    complex boundary corresponds to multiple holes in the surface or bad shaped
    boundary
    In each boundary, the output is a list of vertex indices that respects
    the order of mesh edges i.e. travelling direction along the edges of the
    boundary
    is the same as in edges of the mesh
    :param edges_bound: edges from the mesh that constitue the boundary but are
    not ordered properly
    :param mesh_edges: edges of the mesh that will serve to define the
    travelling direction of the output boundary path
    :return: list of boundaries that are themself lists of vertices
    """
    graph = nx.from_edgelist(edges_bound)

    # import matplotlib.pyplot as plt
    # plt.subplot(111)
    # nx.draw(graph, with_labels=True, font_weight='bold')
    # plt.show()
    def search_edge(ed, edges):
        pos = list()
        for ind, e in enumerate(edges):
            if e[0] == ed[0] and e[1] == ed[1]:
                pos.append(ind)
        return pos

    sub_graphs = [graph.subgraph(c).copy()
                  for c in nx.connected_components(graph)]
    sub_graphs_path = list()
    for sub_graph in sub_graphs:
        # print(len(sub_graph.nodes))
        # plt.subplot(111)
        # nx.draw(sub_graph, with_labels=True, font_weight='bold')
        # plt.show()
        sub_graph_nodes = list(sub_graph.nodes)
        paths = sorted(
            list(
                nx.shortest_simple_paths(
                    sub_graph,
                    source=sub_graph_nodes[0],
                    target=sub_graph_nodes[1]
                )
            ),
            key=len,
            reverse=True,
        )
        longest_path = paths[0]
        if len(longest_path) < len(sub_graph.nodes):
            longest_path = paths[0][1:]
            shortest_path = paths[-1]
            shortest_path.reverse()
            longest_path.extend(shortest_path[1:])
        # print(len(longest_path))
        # print(len(paths))
        # print('---')
        # for p in paths:
        #     print('-'+str(len(p)))
        first_edge = longest_path[0:3]
        pos = search_edge(first_edge, mesh_edges)
        if len(pos) == 0:
            longest_path.reverse()
        sub_graphs_path.append(longest_path)

    return sorted(sub_graphs_path, key=len)


def ismember(ar1, ar2):
    (uni, u_inds) = np.unique(ar1, False, True)
    inds = np.in1d(uni, ar2)
    return np.reshape(inds[u_inds], ar1.shape)


def list_count(liste):
    """
    count the occurrences of all items in a list and return a dictionary
    that is of the form {nb_occurence:list_item}, which is the opposite of
    standard implementation usually found on the web
    -----------------------------------------------------------------
    in python >= 2.7, collections may be used, see example below
    >> from collections import Counter
    >> z = ['blue', 'red', 'blue', 'yellow', 'blue', 'red']
    >> Counter(z)
    Counter({'blue': 3, 'red': 2, 'yellow': 1})
    :param liste:
    :return:
    """
    return dict((liste.count(it), it) for it in liste)


def mesh_boundary(mesh, verbose=False):
    """
    compute borders of a mesh
    :param mesh:
    :return:
    """
    groups = grouping.group_rows(mesh.edges_sorted, require_count=1)
    # vertex_boundary = np.unique(open_mesh.edges_sorted[groups])
    edges_boundary = mesh.edges_sorted[groups]
    if len(edges_boundary) == 0:
        if verbose:
            print("No holes in the surface !!!!")
        return np.array([])
    else:
        return edges_to_boundary(edges_boundary, mesh.edges)


def remove_mesh_boundary_faces(mesh, face_vertex_number=1):
    """
    remove from the mesh the faces and that have face_vertex_number number of
    vertices or more on the boundary of the mesh
    If face_vertex_number = 3 the faces with all 3 vertices on the boundary
    of the mesh will be removed.
    If face_vertex_number = 2 the faces with 2 or 3 vertices on the boundary
    of the mesh will be removed.
    If face_vertex_number = 1 the faces with 1 or 2 or 3 vertices on the
    boundary will be removed.
    The cases face_vertex_number = 0 or >= 3 does not make any sense, so an
    error is raised.
    :param mesh:
    :param face_vertex_number: number of vertices per face lying on the
    boundary of the mesh above which the face has to be removed,
    acceptable values are 1, 2 or 3
    :return: trimesh object
    """
    if face_vertex_number == 0 or face_vertex_number > 3:
        raise NameError(
            "Invalid parameter value: face_vertex_number value "
            "should be 1, 2, or 3"
        )

    open_mesh_boundary = mesh_boundary(mesh)

    boundary_faces = ismember(mesh.faces, open_mesh_boundary[0])
    # compute the mask of faces to keep
    # faces for which face_vertex_number or more vertices are on the boundary
    # are excluded from the mask.
    # i.e. faces with 3-face_vertex_number vertices on the boundary are kept
    face_mask = np.logical_not(np.sum(boundary_faces, 1) >= face_vertex_number)
    out_mesh = mesh.copy()
    out_mesh.update_faces(face_mask)
    out_mesh.remove_unreferenced_vertices()
    return out_mesh


def sub_cut_mesh(mesh, atex, val):
    poly = mesh.faces
    vert = mesh.vertices
    tex_val_indices = np.where(atex == val)[0]
    inds = ismember(poly, tex_val_indices)
    poly_set = poly[inds[:, 0] & inds[:, 1] & inds[:, 2], :]
    (uni, inds) = np.unique(poly_set, False, True)
    submesh = trimesh.Trimesh(
        faces=np.reshape(inds, poly_set.shape),
        vertices=vert[uni, :],
        process=False
    )
    return submesh, tex_val_indices


def texture_boundary(mesh, atex, val):
    """
    compute indexes that are the boundary of a region defined by value
    in a texture
    :param mesh:
    :param atex:
    :param val:
    :return:
    """
    # see mesh.facets_boundary()
    tex_val_indices = np.where(atex == val)[0]
    if not tex_val_indices.size:
        print("no value " + str(val) + " in the input texture!!")
        return list()
    else:

        bound_verts = texture_boundary_vertices(
            atex, val, mesh.vertex_neighbors)
        # select the edges that are on the boundary in the polygons
        u_edges = mesh.edges_unique
        u_edges_count = np.zeros_like(u_edges)
        for v in bound_verts:
            u_edges_count[u_edges == v] += 1
        edges_boundary = u_edges[np.sum(u_edges_count, 1) == 2]
        return edges_to_boundary(edges_boundary, mesh.edges)
        # old method with the wrong adjacency matrix
        # adja_tmp = graph.edges_to_coo(mesh.edges,
        #                               data=np.ones(len(mesh.edges),
        #                                            dtype=np.int8))
        # adja = sparse.triu(adja_tmp) + sparse.tril(adja_tmp).transpose()
        # r = sparse.extract.find(adja)
        # inr0 = []
        # inr1 = []
        # for v in bound_verts:
        #     inr0.extend(np.where(r[0] == v)[0])
        #     inr1.extend(np.where(r[1] == v)[0])
        # r[2][inr0] = r[2][inr0] + 1
        # r[2][inr1] = r[2][inr1] + 1
        # li = r[0][np.where(r[2] == 4)]
        # lj = r[1][np.where(r[2] == 4)]
        # edges_boundary = np.vstack([li, lj]).T


def texture_boundary_vertices(atex, val, vertex_neighbors):
    """
    compute indices of vertices that are the boundary of a region defined by
    value in the texture, without any topological or ordering constraint
    :param atex:
    :param val:
    :param vertex_neighbors:
    :return:
    """
    tex_val_indices = np.where(atex == val)[0]
    if not tex_val_indices.size:
        print("no value " + str(val) + " in the input texture!!")
        return list()
    else:
        ####################################################################
        # print( 'the vertices on the boundary have the same texture value
        # (boundary inside the patch)'
        ####################################################################
        """identify the vertices that are on the boundary,
        i.e that have at least one neighbor that has not the same value in the
        texture"""
        bound_verts = list()
        for i in tex_val_indices:
            ne_i = np.array(vertex_neighbors[i])
            # print( ne_i.size
            # print( np.intersect1d_nu(ne_i, tex_val_indices).size
            inters_size = np.intersect1d(ne_i, tex_val_indices).size
            if inters_size != ne_i.size:
                bound_verts.append(i)
        return bound_verts


def k_ring_neighborhood(mesh, index, k=1, adja=None):
    """
    Generate the k-ring neighborhood around a vertex
    :param mesh:
    :param index:
    :param k:
    :param adja: adjacency matrix
    :return: texture is an array whose values are between 0 and k (all the
    i-ring neighborhoods)
    and k+1 else
    """
    if k > 10:
        raise Exception("k is yoo large (k<=10")
    if adja is None:
        adja = adjacency_matrix(mesh)
    nb_vert = mesh.vertices.shape[0]
    texture = np.zeros((nb_vert, 1))
    texture[index] = k + 1
    tmp_adja = sparse.eye(nb_vert)
    for i in range(k):
        previous_indices = np.where(tmp_adja.getcol(index).toarray() > 0)
        tmp_adja = adja * tmp_adja
        indices = np.where(tmp_adja.getcol(index).toarray() > 0)
        texture[np.setdiff1d(indices[0], previous_indices)] = k - i
    texture = k + 1 - texture
    return texture


def adjacency_matrix(mesh):
    """
    Compute the adjacency matrix of the graph corresponding to the input mesh.
    See https://en.wikipedia.org/wiki/Adjacency_matrix
    :param mesh:
    :return: nb_vertex X bn_vertex sparse matrix (scipy.sparse.coo_matrix)
    """
    adja = graph.edges_to_coo(mesh.edges, data=np.ones(
        len(mesh.edges), dtype=np.int64))
    adja = adja + adja.transpose()
    adja[adja > 0] = 1
    return adja



import numpy as np

from slam import texture


def test_update_darray():
    test_texture = texture.TextureND()
    darray = np.zeros((4, 1))
    test_texture.update_darray(darray)
    assert test_texture.shape == darray.shape



from slam import generate_parametric_surfaces as sps
from slam import vertex_voronoi as svv

test_mesh = sps.generate_sphere_random_sampling(10)
acceptable_error = 0.1


def test_vertex_voronoi():
    """
    compare vertex_voronoi.sum() with mesh.area
    :return:
    """

    vert_vor = svv.vertex_voronoi(test_mesh)
    error = vert_vor.sum() - test_mesh.area
    perc_error = 100 * error / test_mesh.area
    assert perc_error < acceptable_error



import unittest
import tempfile
import numpy as np
import trimesh

import slam.io as sio

# UTILITIES


def make_sphere_a():
    """Create a sphere"""
    mesh_a = trimesh.creation.icosphere(subdivisions=1, radius=1.0)
    return mesh_a


class TestIOMethods(unittest.TestCase):

    # Sphere
    sphere_A = make_sphere_a()

    def test_basic(self):
        mesh_a = self.sphere_A.copy()
        mesh_a_save = self.sphere_A.copy()

        fo = tempfile.NamedTemporaryFile(suffix=".gii")

        sio.write_mesh(mesh_a, fo.name)

        # Non modification
        assert (mesh_a.vertices == mesh_a_save.vertices).all()
        assert (mesh_a.faces == mesh_a_save.faces).all()

        mesh_b = sio.load_mesh(fo.name)

        fo.close()

        precision_A = 0.00001

        # Correctness
        assert np.isclose(mesh_a.vertices, mesh_b.vertices, precision_A).all()
        assert np.isclose(mesh_a.faces, mesh_b.faces, precision_A).all()

    def test_load_texture(self):
        test_text = sio.load_texture('examples/data/example_texture_parcel.gii').darray[0]
        self.assertTrue(test_text.size > 0)


if __name__ == "__main__":
    unittest.main()



"""

"""


import numpy as np
import trimesh
from scipy.sparse.linalg import lgmres


import slam.differential_geometry as sdg
import slam.distortion as sdst
import slam.topology as stop


# error tolerance for the lgmres solver
SOLVER_TOL = 1e-6


def spherical_mapping(
    mesh,
    mapping_type="laplacian_eigenvectors",
    dt=0.01,
    nb_it=10,
    authalic_w=1,
    conformal_w=1,
):
    """Estimate a mapping between a 3d watertight triangular mesh and a sphere

    The sphere is first derived from the 3 first eigen vectors of the input
    triangular mesh. Then several mappings are estimated:

    - Laplacian eigenvectors (see [1]_ for details)

    - Conformal mapping (see [2]_ for details)

    - Authalic mapping (see [3]_ and [4]_ for details)

    - Combined mapping (see [3]_ and [4]_ for details)

    Parameters:
    ----------
    mesh: Trimesh object
     Input topologically spherical triangular mesh to be mapped onto a sphere

    mapping_type: string
        mapping method, possible options are:
        'laplacian_eigenvectors', 'conformal','authalic' or 'combined'

    dt: Float
        discretization step

    nb_it: Int
        number of iterations

    authalic_w: Float
        weight of the authalic constraint for the 'combined' method

    conformal_w: Float
        weight of the conformal constraint for the 'combined' method

    Returns:
    --------
    Trimesh.Mesh
        Spherical representation of the input mesh, having the same
        adjacency (faces, edges, vertex indexing) as the input mesh.

    References:
    ----------
    .. 1_ J. Lefèvre and G. Auzias, "Spherical Parameterization for
    Genus Zero Surfaces Using Laplace-Beltrami Eigenfunctions,"
    in 2nd Conference on Geometric Science of Information, GSI, 2015,
    121–29, https://doi.org/10.1007/978-3-319-25040-3_14.

    .. 2_ Desbrun, M., Meyer, M., & Alliez, P., "Intrinsic parameterizations
    of surface meshes", Computer Graphics Forum, 21(3), 2002,
    209–218. https://doi.org/10.1111/1467-8659.00580

    .. 3_ Rachel a Yotter, Paul M. Thompson, and Christian Gaser,
    “Algorithms to Improve the Reparameterization of Spherical Mappings
    of Brain Surface Meshes.,” Journal of Neuroimaging 21,
    no. 2 (April 2011): e134-47,
    https://doi.org/10.1111/j.1552-6569.2010.00484.x.

    .. 4_ Ilja Friedel, Peter Schröder, and Mathieu Desbrun,
    “Unconstrained Spherical Parameterization”,
    Journal of Graphics, GPU, and Game Tools 12, no. 1 (2007): 17–26.
    """

    # create a sphere from topologically spherical mesh using first three
    # eigenvectors (empirically works)
    sph_vert = sdg.mesh_laplacian_eigenvectors(mesh, nb_vectors=3)
    norm_sph_vert = np.sqrt(np.sum(sph_vert * sph_vert, 1))
    sphere_vertices = sph_vert / np.tile(norm_sph_vert, (3, 1)).T

    if mapping_type == "laplacian_eigenvectors":
        return trimesh.Trimesh(
            faces=mesh.faces,
            vertices=sphere_vertices,
            metadata=mesh.metadata,
            process=False,
        )

    if mapping_type == "conformal":
        L, B = sdg.compute_mesh_laplacian(mesh, lap_type="conformal")

    if mapping_type == "authalic":
        L, B = sdg.compute_mesh_laplacian(mesh, lap_type="authalic")

    if mapping_type == "combined":
        Lconf, Bconf = sdg.compute_mesh_laplacian(mesh, lap_type="conformal")
        Laut, Baut = sdg.compute_mesh_laplacian(mesh, lap_type="authalic")
        L = conformal_w * Lconf + authalic_w * Laut

    # refine the spherical mapping by minimizing the energy
    evol = list()
    for it in range(nb_it):
        sphere_vertices = sphere_vertices - dt * L.dot(sphere_vertices)
        norm_sph_vert = np.sqrt(np.sum(sphere_vertices * sphere_vertices, 1))
        sphere_vertices = sphere_vertices / np.tile(norm_sph_vert, (3, 1)).T
        if it % 10 == 0:
            sph = trimesh.Trimesh(
                faces=mesh.faces, vertices=sphere_vertices, process=False
            )
            angle_diff = sdst.angle_difference(sph, mesh)
            area_diff = sdst.area_difference(sph, mesh)
            edge_diff = sdst.edge_length_difference(sph, mesh)
            evol.append(
                [
                    np.sum(np.abs(angle_diff.flatten())),
                    np.sum(np.abs(area_diff.flatten())),
                    np.sum(np.abs(edge_diff.flatten())),
                ]
            )

    return (
        trimesh.Trimesh(
            faces=mesh.faces,
            vertices=sphere_vertices,
            metadata=mesh.metadata,
            process=False,
        ),
        evol,
    )


def disk_conformal_mapping(
    mesh, lap_type="conformal", boundary=None, boundary_coords=None
):
    """
    Computes comformal mapping of a mesh to a disk, see the following
    references:
    Ulrich Pinkall and Konrad Polthier, “Computing Discrete Minimal
    Surfaces and Their Conjugates,” Experimental Mathematics, 1993, 1–33.
    and
    Mathieu Desbrun, Mark Meyer, and Pierre Alliez, “Intrinsic
    Parameterizations of Surface Meshes,” Computer Graphics Forum 21,
    no. 3 (2002): 209–18,
    https://doi.org/10.1111/1467-8659.00580.
    :param mesh: a trimesh object
    :param lap_type: type of mesh Laplacian to be used,
    see the function differential_geometry/compute_mesh_weights for more
    informations
    :param boundary: boundary of the mesh, resulting from the function
    topology/mesh_boundary
    :param boundary_coords: coordindates of the boundary vertices on the
    output disk, if None then uniform sampling
    :return: a trimesh object, planar disk representation of the input mesh
    """
    if boundary is None:
        boundary_t = stop.mesh_boundary(mesh)
        boundary = boundary_t[0]
    boundary = np.array(boundary)
    if boundary_coords is None:
        p = boundary.size
        t = np.arange(0, 2 * np.pi, (2 * np.pi / p))
        boundary_coords = np.array([np.cos(t), np.sin(t)])
    L, LB = sdg.compute_mesh_laplacian(mesh, lap_type=lap_type)
    Nv = len(mesh.vertices)  # np.array(mesh.vertex()).shape[0]
    print("Boundary Size:", boundary.shape)
    print("Laplacian Size:", L.shape)
    for i in boundary:
        L[i, :] = 0
        L[i, i] = 1
    L = L.tocsr()

    Rx = np.zeros(Nv)
    Ry = np.zeros(Nv)
    Rx[boundary] = boundary_coords[0, :]
    Ry[boundary] = boundary_coords[1, :]

    x, info = lgmres(L, Rx, rtol=SOLVER_TOL)
    y, info = lgmres(L, Ry, rtol=SOLVER_TOL)
    z = np.zeros(Nv)

    return trimesh.Trimesh(
        faces=mesh.faces,
        vertices=np.array([x, y, z]).T,
        metadata=mesh.metadata,
        process=False,
    )


def moebius_transformation(a, b, c, d, plane_mesh):
    """
    see https://en.wikipedia.org/wiki/M%C3%B6bius_transformation
    :param a:Complex
    :param b:Complex
    :param c:Complex
    :param d:Complex
    :param plane_mesh: trimesh mesh
    :return:
    """
    array_complex = plane_mesh.vertices[:,
                                        0] + 1.0j * plane_mesh.vertices[:, 1]
    numerator = (a * array_complex) + b
    denominator = (c * array_complex) + d

    transformed_complex_plane = numerator / denominator
    transformed_vertices = np.array(
        [
            transformed_complex_plane.real,
            transformed_complex_plane.imag,
            plane_mesh.vertices[:, 2],
        ]
    ).T.copy()
    transformed_plane_mesh = trimesh.Trimesh(
        vertices=transformed_vertices,
        faces=plane_mesh.faces.copy(),
        process=False
    )
    return transformed_plane_mesh


def stereo_projection(sphere_mesh, h=None, invert=True):
    """
    Computes the stereographic projection from the unit sphere (center = 0,
    radius = 1) onto the horizontal plane which 3rd coordinate is h of the
    vertices given
    :param sphere_mesh: trimesh spherical mesh to be projected onto the plane
    :param h: 3rd coordinate of the projection plane
    :param invert: Boolean value to invert output mesh faces orientation
    upwards
    :return: trimesh planar mesh, 3rd coordinate is equal to h
    """
    vertices = sphere_mesh.vertices.copy()
    if h is None:
        h = -1
    for ind, vert in enumerate(vertices):
        vertices[ind, 0] = (-h + 1) * vert[0] / (1 - vert[2])
        vertices[ind, 1] = (-h + 1) * vert[1] / (1 - vert[2])
        vertices[ind, 2] = h

    plane_mesh = trimesh.Trimesh(
        vertices=vertices, faces=sphere_mesh.faces.copy(), process=False
    )
    if invert:
        plane_mesh.invert()
    return plane_mesh


def inverse_stereo_projection(plane_mesh, h=None, invert=True):
    """
    Computes the inverse stereograhic projection from an horizontal plane onto
    the unit sphere (center = 0, radius = 1)
    :param plane_mesh: trimesh planar mesh to be inverse projected onto the
    sphere
    :param h: 3rd coordinate of the projection plane
    :param invert: Boolean value to invert input mesh faces orientation upwards
    to be consistent with stereo_projection
    :return: trimesh unit sphere from inverse projected plane_mesh
    """
    if invert:
        plane_mesh.invert()
    vertices = plane_mesh.vertices.copy()
    if h is None:
        h = vertices[0, 2]
    for ind, vert in enumerate(vertices):
        denom = (1 - h) ** 2 + vert[0] ** 2 + vert[1] ** 2
        vertices[ind, 2] = (-((1 - h) ** 2) + vert[0]
                            ** 2 + vert[1] ** 2) / denom
        vertices[ind, 1] = 2 * (1 - h) * vert[1] / denom
        vertices[ind, 0] = 2 * (1 - h) * vert[0] / denom

    sphere_mesh = trimesh.Trimesh(
        vertices=vertices, faces=plane_mesh.faces.copy(), process=False
    )
    return sphere_mesh



"""
Tools for remeshing an input mesh
"""


def spherical_interpolation_nearest_neigbhor(
    source_spherical_mesh, target_spherical_mesh, info_to_interpolate
):
    """
    nearest neighbor interpolation between two spheres
    For each vertex of target_spherical_mesh, find the nearest one
    in source_spherical_mesh and use this vertex-level correspondence
    to pick values in info_to_interpolate
    :param source_spherical_mesh: spherical Trimesh object
    :param target_spherical_mesh: spherical Trimesh object
    :param info_to_interpolate: vector with shape[0] equal to the
    number of vertices in source_spherical_mesh
    :return: interpolated info_to_interpolate
    """
    # This line would interpolate the reverse way (each vertex of
    # source_spherical_mesh, find the nearest one in target_spherical_mesh
    # distance, index =
    # target_spherical_mesh.kdtree.query(source_spherical_mesh.vertices)

    # import time
    # t0 = time.time()

    # the use of kdtree from trimesh is ~100x faster than the loop hereafter
    distance, index = source_spherical_mesh.kdtree.query(
        target_spherical_mesh.vertices)

    # t1 = time.time()
    # source_vertex_number = source_spherical_mesh.vertices.shape[0]
    # nn_corresp = []
    # for v in target_spherical_mesh.vertices:
    #     nn_tmp = np.argmin(np.sum(np.square(
    #     np.tile(v, (source_vertex_number, 1))
    #     - source_spherical_mesh.vertices), 1))
    #     nn_corresp.append(nn_tmp)
    # t2 = time.time()
    # print('with kdtree :'+str(t1-t0))
    # print('with loop :'+str(t2-t1))

    return info_to_interpolate[index]



import unittest
import numpy as np
import trimesh

from slam.surface_profiling import (
    surface_profiling_vert,
    compute_profiles_sampling_points,
    select_points_orientation,
    radial_sort,
    compute_profile_coord_x_y,
    compute_profile_barycentric_para,
    compute_profile_texture_barycentric,
    second_round_profiling_vert,
    vert2poly_indices,
    get_texture_value_on_profile,
    cortical_surface_profiling
)
from slam.texture import TextureND


def make_sphere(radius=1.0, subdivisions=1):
    """Create a sphere mesh for testing."""
    return trimesh.creation.icosphere(subdivisions=subdivisions, radius=radius)


class TestSurfaceProfiling(unittest.TestCase):

    def setUp(self):
        self.mesh = make_sphere(radius=1.0)
        self.vertex_idx = np.argmax(self.mesh.vertices[:, 1])
        self.vertex = self.mesh.vertices[self.vertex_idx]
        self.normal = self.mesh.vertex_normals[self.vertex_idx]

    def test_cortical_surface_profiling(self):
        """Test cortical surface profiling functionality.

        Verifies the computation of surface profiles on a unit sphere,
        checking the output shapes and expected coordinates.
        """
        rot_angle = 90.0  # Simpler angle for verification
        r_step = 0.1
        max_samples = 3

        profile_x, profile_y = cortical_surface_profiling(
            self.mesh, rot_angle, r_step, max_samples
        )

        # Number of profiles per vertex
        expected_profiles = int(360 / rot_angle)
        n_vertices = len(self.mesh.vertices)
        expected_shape = (n_vertices, expected_profiles, max_samples)
        self.assertEqual(profile_x.shape, expected_shape)
        self.assertEqual(profile_y.shape, expected_shape)

        # Test specific values for the top vertex (0,1,0)
        top_vertex_idx = np.argmax(self.mesh.vertices[:, 1])
        # For a unit sphere, x coordinates must be roughly r_step multiples
        expected_x_steps = np.array([0.1, 0.2, 0.3])  # r_step multiples
        np.testing.assert_array_almost_equal(
            profile_x[top_vertex_idx, 0, :], expected_x_steps, decimal=1
        )

        # SImilar idea to expected_x_steps, but for y, starting from 0
        # Bit hacky, but should be reproducible
        expected_y = np.array([-0.03, -0.06, -0.09])
        np.testing.assert_array_almost_equal(
            profile_y[top_vertex_idx, 0, :], expected_y, decimal=2
        )

    def test_surface_profiling_vert(self):
        """Test profiling for a single vertex.

        Verifies the computation of profiles around a single vertex,
        checking output shapes and dimensions.
        TODO: Check the actual profile output
        """
        init_rot_dir = np.array([1, 1, 1]) - self.vertex
        rot_angle = 90.0
        r_step = 1.0
        max_samples = 1

        profiles = surface_profiling_vert(
            self.vertex,
            self.normal,
            init_rot_dir,
            rot_angle,
            r_step,
            max_samples,
            self.mesh,
        )

        # Check output shape
        expected_profiles = int(360 / rot_angle)

        self.assertEqual(profiles.shape[0], expected_profiles)
        self.assertEqual(profiles.shape[1], max_samples)
        self.assertEqual(profiles.shape[2], 3)  # 3D coordinates

    def test_compute_profiles_sampling_points(self):
        """Test computation of profile sampling points.

        Verifies the generation of sampling points along profiles,
        checking output dimensions and point structure.
        Only check the shape, not the content, could
        be improved
        """
        points_intersect = np.array([[0, 0, 1], [0, 1, 0], [1, 0, 0]])
        origin = np.array([0, 0, 0])
        max_samples = 5
        r_step = 0.2

        profile_points = compute_profiles_sampling_points(
            points_intersect, origin, max_samples, r_step
        )

        self.assertEqual(len(profile_points), max_samples)
        self.assertTrue(all(len(point) == 3 for point in profile_points))

    def test_radial_sort(self):
        """Test radial sorting of points around an origin.

        Verifies that points are correctly sorted radially around an origin
        point, maintaining their distances while organizing them in angular
        order.
        """
        points = np.array([[1, 0, 0], [0, 1, 0], [-1, 0, 0], [0, -1, 0]])
        origin = np.array([0, 0, 0])
        normal = np.array([0, 0, 1])

        sorted_points, sorted_indices = radial_sort(points, origin, normal)

        self.assertEqual(len(sorted_points), len(points))
        self.assertEqual(len(sorted_indices), len(points))
        # Points should maintain their distance from origin
        original_distances = np.linalg.norm(points - origin, axis=1)
        sorted_distances = np.linalg.norm(sorted_points - origin, axis=1)
        np.testing.assert_array_almost_equal(
            sorted_distances,
            original_distances
        )

    def test_compute_profile_coord_x_y(self):
        """Test computation of profile coordinates in x-y plane.

        Verifies the projection and computation of profile coordinates
        onto a tangent plane defined by origin and normal vector.
        """
        profiles = np.array(
            [
                [[1, 0, 0], [2, 0, 0]],  # First profile
                [[0, 1, 0], [0, 2, 0]],  # Second profile
            ]
        )
        origin = np.array([0, 0, 0])
        normal = np.array([0, 0, 1])

        x, y = compute_profile_coord_x_y(profiles, origin, normal)

        # Check shapes
        self.assertEqual(x.shape, (2, 2))
        self.assertEqual(y.shape, (2, 2))

        # For these profiles, x should match the distance from origin in the
        # tangent plane
        np.testing.assert_array_almost_equal(x[0], [1, 2])  # First profile
        np.testing.assert_array_almost_equal(x[1], [1, 2])  # Second profile

        # y should be 0 as all points are in the tangent plane
        np.testing.assert_array_almost_equal(y, np.zeros((2, 2)))

    def test_second_round_profiling_vert(self):
        """Test second round of vertex profiling.

        Verifies the extended profiling computation that includes face indices
        and additional geometric information.
        """
        init_rot_dir = np.array([1, 1, 1]) - self.vertex
        rot_angle = 90.0
        r_step = 0.1
        max_samples = 3
        mesh_face_index = np.arange(len(self.mesh.faces))

        profile_points, face_indices = second_round_profiling_vert(
            self.vertex,
            self.normal,
            init_rot_dir,
            rot_angle,
            r_step,
            max_samples,
            self.mesh,
            mesh_face_index,
        )

        # Check output shapes
        expected_profiles = int(360 / rot_angle)
        self.assertEqual(profile_points.shape[0], expected_profiles)
        self.assertEqual(profile_points.shape[1], max_samples)
        self.assertEqual(profile_points.shape[2], 3)  # [p1, p2, sample_point]
        self.assertEqual(profile_points.shape[3], 3)  # 3D coordinates

        # Check face indices shape
        self.assertEqual(face_indices.shape[0], expected_profiles)
        self.assertEqual(face_indices.shape[1], max_samples)

    def test_vert2poly_indices(self):
        """Test vertex to polygon index mapping.

        Verifies the correct identification of polygons containing
        specified vertices in the mesh.
        """
        poly_array = np.array([[0, 1, 2], [1, 2, 3], [2, 3, 4]])
        vertex_array = np.array([3, 4])

        result = vert2poly_indices(vertex_array, poly_array)

        # Vertices 3 and 4 should be in polygons 1 and 2
        expected = np.array([1, 2])
        np.testing.assert_array_equal(np.sort(result), expected)

    def test_get_texture_value_on_profile(self):
        """Test retrieval of texture values along profiles.

        Verifies the correct sampling of texture values at profile points
        using mesh and texture information.

        Could probably design a better test, with texture being
        not all ones
        """
        texture = np.ones((len(self.mesh.vertices),))

        # Create sample profile points
        # 1 vertex, 2 profiles, 2 samples
        profile_samples = np.zeros((1, 2, 2, 3, 3))
        profile_samples_fid = np.zeros(
            (1, 2, 2), dtype=int
        )  # Corresponding face indices

        # Create proper TextureND object instead of mock
        texture_obj = TextureND(darray=texture)

        result = get_texture_value_on_profile(
            texture_obj, self.mesh, profile_samples, profile_samples_fid
        )

        # Should match input profile shape
        self.assertEqual(result.shape, (1, 2, 2))

        # should be all ones, as the original texture is all ones
        np.testing.assert_array_equal(result, np.ones((1, 2, 2)))

    def test_compute_profile_barycentric_para(self):
        """Test computation of barycentric coordinates for profile points.

        Verifies the calculation of barycentric coordinates for profile points
        within their respective triangles.
        """
        # 1 vertex, 2 profiles, 2 samples
        profile_points = np.zeros((1, 2, 2, 3, 3))
        triangle_id = np.zeros((1, 2, 2), dtype=int)

        barycentric = compute_profile_barycentric_para(
            profile_points, self.mesh, triangle_id
        )

        # Check output shape
        self.assertEqual(
            barycentric.shape, (1, 2, 2, 3)
        )  # Last dimension is barycentric coords

    def test_compute_profile_texture_barycentric(self):
        """Test computation of texture values using barycentric coordinates.

        Verifies the interpolation of texture values at profile points
        using barycentric coordinates within triangles.
        """
        texture = np.ones(len(self.mesh.vertices))
        # 1 vertex, 2 profiles, 2 samples
        triangle_id = np.zeros((1, 2, 2), dtype=int)
        # Equal barycentric coordinates
        barycentric_coord = np.ones((1, 2, 2, 3)) / 3

        result = compute_profile_texture_barycentric(
            texture, self.mesh, triangle_id, barycentric_coord
        )

        # Check output shape
        self.assertEqual(result.shape, (1, 2, 2))

        # For uniform texture and equal barycentric coordinates, result should
        # be 1
        np.testing.assert_array_almost_equal(result, np.ones((1, 2, 2)))

    def test_select_points_orientation(self):
        """Test selection and orientation of intersection points.

        Verifies the correct selection and ordering of intersection points
        with a simple example
        """
        intersect_points = np.array(
            [
                [[0, 0, 0], [1, 0, 0]],  # Line 1
                [[0, 0, 0], [0, 1, 0]],  # Line 2
                [[0, 0, 0], [-1, 0, 0]],  # Line 3 (opposite to Line 1)
                [[0, 0, 0], [0, -1, 0]],  # Line 4 (opposite to Line 2)
            ]
        )

        origin = np.array([0, 0, 0])
        normal = np.array([0, 0, 1])
        # Direction vector pointing towards positive x
        r_alpha = np.array([1, 0, 0])

        # TODO: check all outputs
        orient_points, _, lines_indices = select_points_orientation(
            intersect_points, r_alpha, origin, normal
        )

        # Check that we get points in positive x direction
        self.assertTrue(
            np.all(orient_points[:, 0] >= 0)
        )  # All x coordinates should be positive

        # Check shapes
        self.assertEqual(orient_points.shape[1], 3)  # Each point should be 3D

        # Check that points are ordered by distance from origin
        distances = np.linalg.norm(orient_points - origin, axis=1)
        self.assertTrue(
            np.all(np.diff(distances) >= 0)
        )  # Distances should be non-decreasing

        # Check that indices are valid
        self.assertTrue(np.all(lines_indices >= 0))
        self.assertTrue(np.all(lines_indices < len(intersect_points)))


if __name__ == "__main__":
    unittest.main()



import slam.distortion as sdst
import numpy as np
import unittest
import trimesh

# UTILITIES


def make_sphere(radius=1):
    """Create a sphere"""
    mesh_a = trimesh.creation.icosphere(subdivisions=1, radius=radius)
    return mesh_a


class TestDistortionMethods(unittest.TestCase):

    # Spheres
    sphere_r1 = make_sphere(1)
    sphere_r2 = make_sphere(2)

    def test_basic(self):
        mesh_a = self.sphere_r1.copy()
        mesh_a_save = self.sphere_r1.copy()

        mesh_b = self.sphere_r2.copy()
        mesh_b_save = self.sphere_r2.copy()

        # tmp = sdst.angle_difference(mesh_a, mesh_b)

        assert (mesh_a.vertices == mesh_a_save.vertices).all()
        assert (mesh_a.faces == mesh_a_save.faces).all()

        assert (mesh_b.vertices == mesh_b_save.vertices).all()
        assert (mesh_b.faces == mesh_b_save.faces).all()

        # tmp = sdst.area_difference(mesh_a, mesh_b)

        assert (mesh_a.vertices == mesh_a_save.vertices).all()
        assert (mesh_a.faces == mesh_a_save.faces).all()

        assert (mesh_b.vertices == mesh_b_save.vertices).all()
        assert (mesh_b.faces == mesh_b_save.faces).all()

        # tmp = sdst.edge_length_difference(mesh_a, mesh_b)

        assert (mesh_a.vertices == mesh_a_save.vertices).all()
        assert (mesh_a.faces == mesh_a_save.faces).all()

        assert (mesh_b.vertices == mesh_b_save.vertices).all()
        assert (mesh_b.faces == mesh_b_save.faces).all()

    def test_correctness_area(self):

        precisionA = 0.001

        n_sub = 5

        mesh_a = trimesh.creation.icosphere(subdivisions=n_sub, radius=1)
        mesh_b = trimesh.creation.icosphere(subdivisions=n_sub, radius=2)

        area_diff_estim = sdst.area_difference(mesh_a, mesh_b)

        area_diff_estim = sum(abs(area_diff_estim))

        area_diff_analytical = 4 * np.pi * 2 * 2 - 4 * np.pi * 1 * 1

        assert np.isclose(area_diff_estim, area_diff_analytical, precisionA)

    def test_correctness_angle(self):

        precisionA = 0.001

        n_sub = 5

        mesh_a = trimesh.creation.icosphere(subdivisions=n_sub, radius=1)
        mesh_b = trimesh.creation.icosphere(subdivisions=n_sub, radius=2)

        angle_diff_estim = sdst.angle_difference(mesh_a, mesh_b)

        angle_diff_estim = sum(sum(abs(angle_diff_estim)))

        angle_diff_analytical = 0

        assert np.isclose(angle_diff_estim, angle_diff_analytical, precisionA)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import unittest
import trimesh
import slam.curvature as scurv
import slam.generate_parametric_surfaces as sgps
import slam.utils as ut

# UTILITIES


def make_sphere():
    """Create a sphere"""
    mesh_a = trimesh.creation.icosphere(subdivisions=1, radius=1.0)
    return mesh_a


def make_box():
    """Create a box"""
    mesh_a = trimesh.creation.box((1, 1, 1))
    return mesh_a


class TestCurvatureMethods(unittest.TestCase):

    # Sphere
    sphere_A = make_sphere()

    # Box
    box_A = make_box()

    def test_basic(self):
        mesh_a = self.sphere_A.copy()
        mesh_a_save = self.sphere_A.copy()

        cur, d1, d2 = scurv.curvatures_and_derivatives(mesh_a)

        # Non modification
        assert (mesh_a.vertices == mesh_a_save.vertices).all()
        assert (mesh_a.faces == mesh_a_save.faces).all()

    def test_local_basis(self):
        precision = 1e-12
        # First test, generic situation
        normal1 = 1 / np.sqrt(3) * np.array([[1], [1], [1]])
        e1 = 1 / np.sqrt(6) * np.array([[2], [-1], [-1]])
        e2 = 1 / np.sqrt(2) * np.array([[0], [1], [-1]])
        res = scurv.determine_local_basis(normal1, precision)
        assert np.isclose(res[0], e1, precision).all()
        assert np.isclose(res[1], e2, precision).all()

        # Second test, when norm(vec1) < tol in determine_local_basis
        normal1 = np.array([[1], [0], [0]])
        e1 = np.array([[0], [1], [0]])
        e2 = np.array([[0], [0], [1]])
        res = scurv.determine_local_basis(normal1, precision)
        assert np.isclose(res[0], e1, precision).all()
        assert np.isclose(res[1], e2, precision).all()

    def test_correctness_curvature_sphere(self):

        # Iterations on radius

        for i in [1, 2, 4]:

            precision_A = 0.000001
            precision_B = 0.000001

            with self.subTest(i):

                # Sphere of radius i
                mesh_a = trimesh.creation.icosphere(
                    subdivisions=1, radius=float(i))

                curv, d1, d2 = scurv.curvatures_and_derivatives(mesh_a)

                mean_curv = 0.5 * (curv[0, :] + curv[1, :])

                gauss_curv = curv[0, :] * curv[1, :]

                shape = mean_curv.shape

                # The mean curvature in every point is 1/radius
                analytical_mean = np.full(shape, 1 / i)

                # The gaussian curvature in every point is 1/radius**2
                analytical_gauss = np.full(shape, 1 / i**2)

                assert np.isclose(
                    mean_curv, analytical_mean, precision_A).all()
                assert np.isclose(
                    gauss_curv,
                    analytical_gauss,
                    precision_B).all()

        # Iterations on subdivisions

        for i in [1, 2, 3]:

            precision_A = 0.000001
            precision_B = 0.000001

            with self.subTest(i):

                # Sphere of radius 2, i subdivisions
                mesh_a = trimesh.creation.icosphere(subdivisions=i, radius=2)

                curv, d1, d2 = scurv.curvatures_and_derivatives(mesh_a)

                mean_curv = 0.5 * (curv[0, :] + curv[1, :])
                gauss_curv = curv[0, :] * curv[1, :]

                shape = mean_curv.shape

                # The curvature in every point is 1/radius
                analytical_mean = np.full(shape, 1 / 2)

                # The curvature in every point is 1/radius**2
                analytical_gauss = np.full(shape, 1 / 4)

                assert np.isclose(
                    mean_curv, analytical_mean, precision_A).all()
                assert np.isclose(
                    analytical_gauss,
                    analytical_gauss,
                    precision_B).all()

    # @unittest.skip
    def test_correctness_curvature_quadric(self):

        K = [1, 1]
        quadric = sgps.generate_quadric(
            K,
            nstep=[20, 20],
            ax=3,
            ay=1,
            random_sampling=True,
            ratio=0.3,
            random_distribution_type="gamma",
        )

        # Computation of estimated curvatures

        p_curv, d1, d2 = scurv.curvatures_and_derivatives(quadric)

        k1_estim, k2_estim = p_curv[0, :], p_curv[1, :]

        # k_gauss_estim = k1_estim * k2_estim

        k_mean_estim = 0.5 * (k1_estim + k2_estim)

        # Computation of analytical curvatures

        k_mean_analytic = sgps.quadric_curv_mean(K)(
            np.array(quadric.vertices[:, 0]), np.array(quadric.vertices[:, 1])
        )

        k_gauss_analytic = sgps.quadric_curv_gauss(K)(
            np.array(quadric.vertices[:, 0]), np.array(quadric.vertices[:, 1])
        )

        k1_analytic = np.zeros((len(k_mean_analytic)))
        k2_analytic = np.zeros((len(k_mean_analytic)))

        for i in range(len(k_mean_analytic)):
            a, b = np.roots((1, -2 * k_mean_analytic[i], k_gauss_analytic[i]))
            k1_analytic[i] = min(a, b)
            k2_analytic[i] = max(a, b)

        # /// STATS

        k_mean_relative_change = abs(
            (k_mean_analytic - k_mean_estim) / k_mean_analytic)
        k_mean_absolute_change = abs((k_mean_analytic - k_mean_estim))

        k1_relative_change = abs((k1_analytic - k1_estim) / k1_analytic)
        # k1_absolute_change = abs((k1_analytic - k1_estim))

        a = []
        a += [
            [
                "K_MEAN",
                "mean",
                "relative change",
                np.mean(k_mean_relative_change * 100),
                "%",
            ],
            (
                "K_MEAN",
                "std",
                "relative change",
                np.std(k_mean_relative_change * 100),
                "%",
            ),
            (
                "K_MEAN",
                "max",
                "relative change",
                np.max(k_mean_relative_change * 100),
                "%",
            ),
            ["K_MEAN", "mean", "absolute change",
                np.mean(k_mean_absolute_change)],
            ["K_MEAN", "std", "absolute change",
                np.std(k_mean_absolute_change)],
            ["K_MEAN", "max", "absolute change",
                np.max(k_mean_absolute_change)],
            (" K1", "mean", "relative change",
             np.mean(k1_relative_change * 100), "%"),
            (" K1", "std", "relative change", np.std(
                k1_relative_change * 100), "%"),
            (" K1", "max", "relative change", np.max(
                k1_relative_change * 100), "%"),
            (" K1", "mean", "absolute change",
             np.mean(k_mean_absolute_change)),
            (" K1", "std", "absolute change",
             np.std(k_mean_absolute_change)),
            (" K1", "max", "absolute change",
             np.max(k_mean_absolute_change)),
        ]

        # PRINT STATS
        print("----------------------------------------")
        for i, v in enumerate(a):
            numeric_value = np.round(v[3], decimals=3)
            if i == 6:
                print("----------------------------------------")
            if len(v) > 4:
                print(
                    "{0:10} {1:5} {2:16} {3:2} {4} {5}".format(
                        v[0], v[1], v[2], "=", numeric_value, v[4]
                    )
                )
            else:
                print(
                    "{0:10} {1:5} {2:16} {3:2} {4}".format(
                        v[0], v[1], v[2], "=", numeric_value
                    )
                )
        print("----------------------------------------")

    def test_error_drop_curvature_quadric(self):
        """Asserts the error drops when the mesh resolution is increased"""

        out = []

        for j in range(3):
            K = [1, 1]

            # Increase of the number of points
            quadric = sgps.generate_quadric(
                K,
                nstep=[20 + 10 * j, 20 + 10 * j],
                ax=3,
                ay=1,
                random_sampling=False,
                ratio=0.3,
                random_distribution_type="gamma",
            )

            # Computation of estimated curvatures

            p_curv, d1, d2 = scurv.curvatures_and_derivatives(quadric)

            k1_estim, k2_estim = p_curv[0, :], p_curv[1, :]

            # k_gauss_estim = k1_estim * k2_estim

            k_mean_estim = 0.5 * (k1_estim + k2_estim)

            # Computation of analytical curvatures

            k_mean_analytic = sgps.quadric_curv_mean(K)(
                np.array(quadric.vertices[:, 0]), np.array(
                    quadric.vertices[:, 1])
            )

            k_gauss_analytic = sgps.quadric_curv_gauss(K)(
                np.array(quadric.vertices[:, 0]), np.array(
                    quadric.vertices[:, 1])
            )

            k1_analytic = np.zeros((len(k_mean_analytic)))
            k2_analytic = np.zeros((len(k_mean_analytic)))

            for i in range(len(k_mean_analytic)):
                a, b = np.roots(
                    (1, -2 * k_mean_analytic[i], k_gauss_analytic[i]))
                k1_analytic[i] = min(a, b)
                k2_analytic[i] = max(a, b)

            # /// STATS

            # k_mean_relative_change = abs(
            #     (k_mean_analytic - k_mean_estim) / k_mean_analytic)
            k_mean_absolute_change = abs((k_mean_analytic - k_mean_estim))

            # k1_relative_change = abs((k1_analytic - k1_estim) / k1_analytic)
            # k1_absolute_change = abs((k1_analytic - k1_estim))

            out += [np.round(np.mean(k_mean_absolute_change), 6)]

        # Assert the absolute mean error decreases as we increase the number of
        # points
        assert sorted(out, reverse=True) == out

    def test_error_drop_curvature_sphere(self):

        out = []

        for j in range(3):
            mesh_a = trimesh.creation.icosphere(subdivisions=j + 1, radius=2)
            p_curv, d1, d2 = scurv.curvatures_and_derivatives(mesh_a)
            k1_estim, k2_estim = p_curv
            k_mean_estim = (k1_estim + k2_estim) * 0.5
            # k_gauss_estim = (k1_estim * k2_estim)

            out += [np.round(np.mean(k_mean_estim), 6)]

        assert sorted(out, reverse=True) == out

    def test_correctness_decomposition_sphere(self):

        precisionA = 0.0000001
        precisionB = 0.0000001

        radius = 3

        mesh_a = trimesh.creation.icosphere(subdivisions=2, radius=radius)

        mesh_a_save = mesh_a.copy()

        shapeIndex, curvedness = scurv.curvedness_shapeIndex(mesh_a)

        # Non modification

        assert (mesh_a.vertices == mesh_a_save.vertices).all()
        assert (mesh_a.faces == mesh_a_save.faces).all()

        # Correctness
        # ShapeIndex is 1 or -1 on every vertex
        # Curvedness is 1/radius on every vertex
        assert (
            np.isclose(shapeIndex, 1, precisionA)
            | np.isclose(shapeIndex, -1, precisionA)
        ).all()
        assert np.isclose(curvedness, 1 / radius, precisionB).all()

    def test_correctness_decomposition_quadric(self):

        # Precision on the shapeIndex calculation
        precisionA = 0.000001
        # Precision on the curvedness calculation
        precisionB = 0.000001

        set_of_tests = [
            # Set = [K, correct shapeIndex, correct curvedness]
            # For example, on a quadric generated by K=[1,1], the shapeIndex
            # is +-1 at the center and the curvedness is +-2
            [[1, 1], 1, 2],
            [[-1, -1], 1, 2],
            [[0.5, 0.5], 1, 1],
            [[1, 0], 0.5, np.sqrt(2)],
        ]

        for i in range(len(set_of_tests)):

            current_test = set_of_tests[i]

            K = current_test[0]

            # Correct values of shapeIndex and curvedness on the vertex at the
            # center of the mesh
            correct_shape_index = current_test[1]
            correct_curvedness = current_test[2]

            with self.subTest(i):

                # Generate quadric
                quadric = sgps.generate_quadric(
                    K,
                    nstep=[20, 20],
                    ax=3,
                    ay=1,
                    random_sampling=False,
                    ratio=0.3,
                )

                # Computation of analytical k_mean, k_gauss, k_1 and k_2
                k_mean_analytic = sgps.quadric_curv_mean(K)(
                    np.array(quadric.vertices[:, 0]), np.array(
                        quadric.vertices[:, 1])
                )

                k_gauss_analytic = sgps.quadric_curv_gauss(K)(
                    np.array(quadric.vertices[:, 0]), np.array(
                        quadric.vertices[:, 1])
                )

                k1_analytic = np.zeros((len(k_mean_analytic)))
                k2_analytic = np.zeros((len(k_mean_analytic)))

                for i in range(len(k_mean_analytic)):
                    a, b = np.roots(
                        (1, -2 * k_mean_analytic[i], k_gauss_analytic[i]))
                    k1_analytic[i] = min(a, b)
                    k2_analytic[i] = max(a, b)

                # Decomposition of the curvature
                shapeIndex, curvedness = scurv.decompose_curvature(
                    np.array((k1_analytic, k2_analytic))
                )

                # Find the index of the vertex which is the closest to the
                # center

                def mag(p):
                    """Distance to the center for the point p, not considering
                    the z axis"""
                    x = p[0]
                    y = p[1]
                    return np.sqrt(x * x + y * y)

                min_i = 0
                min_m = mag(quadric.vertices[0])
                for i, v in enumerate(quadric.vertices):
                    mg = mag(v)
                    if mg < min_m:
                        min_i = i
                        min_m = mg

                # Correctness
                computed_shapeIndex = shapeIndex[min_i]
                computed_curvedness = curvedness[min_i]

                assert np.isclose(
                    computed_shapeIndex, correct_shape_index, precisionA
                ) | np.isclose(computed_shapeIndex, -correct_shape_index, precisionA)
                assert np.isclose(
                    computed_curvedness, correct_curvedness, precisionB
                ) | np.isclose(computed_curvedness, -correct_curvedness, precisionB)

    def test_correctness_direction_quadric(self):

        # WARNING: THRESHOLD should be st to 15°, here this value is just to
        # ensure the test is okay
        # NEED to check Rusinkiewicz method and the proper orientation of
        # curvature directions
        THRESHOLD = 10

        # Generate a paraboloid
        K = [1, 0]

        # Increase of the number of points
        quadric = sgps.generate_quadric(
            K,
            nstep=[20, 20],
            ax=3,
            ay=3,
            random_sampling=False,
            ratio=0.3,
            random_distribution_type="gamma",
            equilateral=True,
        )

        # ESTIMATED Principal curvature, Direction1, Direction2
        p_curv_estim, d1_estim, d2_estim = scurv.curvatures_and_derivatives(
            quadric)

        # ANALYTICAL directions
        analytical_directions = sgps.compute_all_principal_directions_3D(
            K, quadric.vertices
        )

        estimated_directions = np.zeros(analytical_directions.shape)
        estimated_directions[:, :, 0] = d1_estim
        estimated_directions[:, :, 1] = d2_estim

        angular_error_0, dotprods = ut.compare_analytic_estimated_directions(
            analytical_directions[:, :, 0], d2_estim, abs=True
        )
        angular_error_0 = 180 * angular_error_0 / np.pi

        # CORRECTNESS DIRECTION 1

        # Number of vertices where the angular error is lower than 20 degrees
        n_low_error = np.sum(angular_error_0 < THRESHOLD)

        # Percentage of vertices where the angular error is lower than 20
        # degrees
        percentage_low_error = n_low_error / angular_error_0.shape[0]

        assert percentage_low_error > 0.75

        # CORRECTNESS DIRECTION 2

        angular_error_1, dotprods = ut.compare_analytic_estimated_directions(
            analytical_directions[:, :, 1], d1_estim, abs=True
        )
        angular_error_1 = 180 * angular_error_1 / np.pi

        # Number of vertices where the angular error is lower than 20 degrees
        n_low_error = np.sum(angular_error_1 < THRESHOLD)

        # Percentage of vertices where the angular error is lower than 20
        # degrees
        percentage_low_error = n_low_error / angular_error_1.shape[0]

        assert percentage_low_error > 0.75


if __name__ == "__main__":

    unittest.main()



import unittest
import numpy as np
from scipy.spatial import Delaunay
import trimesh
import slam.geodesics as sg


TOL = 1e-10


def create_rectangular_grid(rows=3, cols=4):
    """
    Create a rectangular graph
    :param rows:
    :param cols:
    :return:
    """
    x = np.arange(0, rows, 1)
    y = np.arange(0, cols, 1)

    X, Y = np.meshgrid(x, y)
    X = X.flatten()
    Y = Y.flatten()
    Z = np.zeros((len(X),))

    coords = np.vstack((X, Y, Z)).T
    tri = Delaunay(np.vstack((X, Y)).T, qhull_options="QJ Qt Qbb")
    print(tri)

    return trimesh.Trimesh(faces=tri.simplices, vertices=coords, process=False)



class TestGeodesics(unittest.TestCase):

    def test_compute_gdist(self):
        # Check if the distance between two points of a diagonal
        # is bounded between L2 norm and L1 norm
        m = 4
        n = 5
        rectangle = create_rectangular_grid(m, n)
        print(rectangle)
        distance = sg.compute_gdist(rectangle, 0)
        low_bound = np.sqrt((m - 1) ** 2 + (n - 1) ** 2)  # L2 diameter
        up_bound = m - 1 + n - 1  # L1 diameter
        self.assertTrue(np.logical_and(distance[m * n - 1] >= low_bound,
                                       distance[-1] <= up_bound))

    def test_compute_gdist_to_target(self):
        m = 4
        n = 5
        rectangle = create_rectangular_grid(m, n)
        print(rectangle)
        target = 10 #rectangle.vertices.shape[0]-1
        distance = sg.compute_gdist(rectangle, 0)
        distance_to_target = sg.compute_gdist(rectangle, 0, target)
        self.assertTrue(distance[target]==distance_to_target)

    # def test_dijkstra_length(self):
    #     # Check if the max distance to one side of the rectangle
    #     # equals the length of the other side
    #     m = 4
    #     n = 5
    #     rectangle = create_rectangular_grid(m, n)
    #     set_of_points = range(m)  # boundary with y = 0
    #     distance = sg.dijkstra_length(rectangle, set_of_points)
    #     self.assertTrue(np.abs(np.max(distance) - (n - 1) < TOL))
    #
    # def test_gdist_length(self):
    #     # Check if the max distance to one side of the rectangle
    #     # equals the length of the other side
    #     m = 4
    #     n = 5
    #     rectangle = create_rectangular_grid(m, n)
    #     set_of_points = range(m)  # boundary with y = 0
    #     distance = sg.gdist_length(rectangle, set_of_points)
    #     self.assertTrue(np.abs(np.max(distance) - (n - 1) < TOL))


if __name__ == "__main__":
    unittest.main()



import numpy as np


def clamp(val, a, b):
    if val < a:
        return a
    if val > b:
        return b
    return val


def angle(vec1, vec2, abs=False):
    """
    Return the angle between two vectors
    :param vec1:
    :param vec2:
    :return:
    """
    dp = dotprod(vec1, vec2)
    if abs:
        dp = np.abs(dp)
    return np.arccos(clamp(dp, -1, 1))


def dotprod(vec1, vec2):
    """
    Return the normalized dotprod between two vectors
    :param vec1:
    :param vec2:
    :return:
    """
    return np.dot(vec1, vec2) / \
        (np.sqrt(np.dot(vec1, vec1) * np.dot(vec2, vec2)))


def compare_analytic_estimated_directions(
    analytic_directions, estimated_directions, abs=False
):
    """
    Compare the analytic principal directions with a estimated directions
    :param analytic_directions: (n,3) array
    :param estimated_directions: (n,3) array with one direction
    :return: angular error distribution
    """
    n = analytic_directions.shape[0]
    angular_error = np.zeros((n,))
    dotprods = np.zeros((n,))
    for i in range(n):
        angle0 = angle(analytic_directions[i, :],
                       estimated_directions[i, :], abs)
        angular_error[i] = angle0
        dotprods[i] = dotprod(analytic_directions[i, :],
                              estimated_directions[i, :])

    return angular_error, dotprods


def compare_analytic_estimated_directions_min(
    analytic_directions, estimated_directions
):
    """
    Compare the analytic principal directions with a estimated directions
    :param analytic_directions: (n,3) array
    :param estimated_directions: (n,3,2) array with the two directions
    that may be not necessarily ordered
    :return: angular error distribution
    """
    n = analytic_directions.shape[0]
    angular_error = np.zeros((n,))
    dotprods = np.zeros((n, 2))
    for i in range(n):
        angle0 = angle(analytic_directions[i, :],
                       estimated_directions[i, :, 0])
        angle1 = angle(analytic_directions[i, :],
                       estimated_directions[i, :, 1])
        angular_error[i] = min(
            np.mod(
                angle0,
                np.pi / 4),
            np.mod(
                angle1,
                np.pi / 4))
        dotprods[i, 0] = dotprod(
            analytic_directions[i, :], estimated_directions[i, :, 0]
        )
        dotprods[i, 1] = dotprod(
            analytic_directions[i, :], estimated_directions[i, :, 1]
        )
    return angular_error, dotprods


def get_rotate_matrix(rot_axis, angle):
    """
    for a pair of rotation axis and angle, calculate the rotate matrix
    :param rot_axis: rotation axis
    :param angle: float, the rotation angle is a real number rather than degree
    :return: rotate matrix of [3, 3]
    """

    if np.linalg.norm(rot_axis) == 0:
        raise Exception("The axis of rotation cannot be 0.")

    # normalize the rotate axis
    r_n = rot_axis / np.linalg.norm(rot_axis)
    rot_matrix = np.zeros((3, 3), dtype="float32")

    cos_theta = np.cos(angle)
    sin_theta = np.sin(angle)
    x = r_n[0]
    y = r_n[1]
    z = r_n[2]

    rot_matrix[0, 0] = cos_theta + (1 - cos_theta) * np.power(x, 2)
    rot_matrix[0, 1] = (1 - cos_theta) * x * y - sin_theta * z
    rot_matrix[0, 2] = (1 - cos_theta) * x * z + sin_theta * y

    rot_matrix[1, 0] = (1 - cos_theta) * y * x + sin_theta * z
    rot_matrix[1, 1] = cos_theta + (1 - cos_theta) * np.power(y, 2)
    rot_matrix[1, 2] = (1 - cos_theta) * y * z - sin_theta * x

    rot_matrix[2, 0] = (1 - cos_theta) * z * x - sin_theta * y
    rot_matrix[2, 1] = (1 - cos_theta) * z * y + sin_theta * x
    rot_matrix[2, 2] = cos_theta + (1 - cos_theta) * np.power(z, 2)

    return rot_matrix


def project_vector2tangent_plane(v_n, v_p):
    """
    calculate the projection vector of v_p onto tangent plane of v_n
    :param v_n: array of (3,)  float
        normal vector of v
    :param v_p: array of (n, 3) float
        vector projected
    :return: v_t (n, 3) float
        projection result
    """

    if np.linalg.norm(v_n) == 0:
        unitev_n = v_n
    else:
        unitev_n = v_n / np.linalg.norm(v_n)

    coeff_v_pn = np.dot(v_p, unitev_n)

    coeff = coeff_v_pn.reshape([coeff_v_pn.size, 1])

    v_pn = coeff * unitev_n

    v_t = v_p - np.array(v_pn)

    return v_t


def project_vector2vector(v_n, v_p):
    """
    calculate the projection vector of v_p onto v_n,
    v_pn = (v_p dot v_n) / |v_n| * unite vector of (v_n)
    :param v_n: array of (3,)  float
        direction vector
    :param v_p: array of (n, 3) float
        vectors projected
    :return: (n, 3) float
        projection result
    """

    if np.linalg.norm(v_n) == 0:
        unitev_n = v_n
    else:
        unitev_n = v_n / np.linalg.norm(v_n)

    coeff_v_pn = np.dot(v_p, unitev_n)

    coeff = coeff_v_pn.reshape([coeff_v_pn.size, 1])

    v_pn = coeff * unitev_n

    return v_pn
